# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from sqlalchemy import func, or_, String
from sqlalchemy.inspection import inspect as inspecc
from sqlalchemy.orm import joinedload, aliased

from collections.abc import Iterable

import logging
logger = logging.getLogger("DataModel")


def _isCollection(obj):
    """Determine if whether an object is a collection."""
    return isinstance(obj, Iterable) and not isinstance(obj, str)


def _str2bool(string):
    if string.lower() in ("true", "1", "yes"):
        return True
    elif string.lower() in ("false", "0", "no"):
        return False
    raise ValueError("Argument must be either 'true' or 'false'")


class MismatchROError(BaseException):
    """Exception raised when a read-only attribute does not match."""

    pass


class InvalidAttributeError(BaseException):
    """Exception raised when an unknown attribute is to be modified."""

    pass


class MissingRequiredAttributeError(BaseException):
    """Exception raised when a required attribute is not specified."""

    pass


class DataModel:
    """Central class for data structuring.

    Provides standardizes implementation for reading and writing data from and to SQLAlchemy mapped tables.
    """

    class Prop:
        """Property representation.

        Manages property settings.
        """

        def __init__(self, attr, alias=None, flags=None, args=(), kwargs={}, mask=None, target=None, dispname=None,
                     flat=None, func=None, link=None, filter=None, qopt=joinedload, proxy=None, arg_tf=None, match="default",
                     **unknown):
            """
            Initialize Prop.

            Parameters
            ----------
            attr : str
                Name of the attribute to map.
            alias : str, optional
                Alias of the attribute. If set, effectively replaces `attr` to front-facing operations. The default is None.
            flags : str, optional
                Comma separated list of flags. Recognized values are:
                    - ref: The attribute is a relationship that should be dereferenced
                    - call: The attribute is a method that returns the actual value
                    - sort: The attribute is available for sorting. Can also be set with _sortables_
                    - patch: The attribute can be set/updated
                    - init: Like `patch` but only if the current value is `None`
                    - managed: The referenced object(s) are fully managed through this class (enables transparent CUD)
                        The default is None.
                    - match: Use this attribute for matching. Can also be set with _matchables_
                    - hidden: The attribute is not included at any level
            args : Collection, optional
                List of arguments passed to the function if `call` or `func` is set. The default is None.
            kwargs : dict, optional
                Keyword arguments passed to the function if `call` or `func` is set. The default is None.
            mask : str, optional
                Name of the attribute masked by the relationship (i.e. the foreign key column). The default is None.
            target : str, optional
                Name of the foreign attribute to use for sorting and matching. The default is None.
            flat : str, optional
                If set, do not serialize ref into an object, but use its `flat` attribute directly. The default is None.
            func : function, optional
                Call `func` on the attribute. Effectively executes func(attr, *args, **kwargs). The default is None.
            link : str, optional
                Name of the attribute identifying the foreign object. The default is None.
            filter : str, optional
                Type of multi-value filtering to apply. Valid values are 'set' and 'range. The default is None.
            qopt : function, optional
                Type of query optimization to use for the relationship. The default is `joinedload`.
            proxy: str, optional
                Name of the relationship containing the attribute. Usable only for search/filter/match operations.
            arg_tf: function, optional
                Function transform request arguments into the correct format/type. Used by filter and match operations.
            match: str, optional
                Mode used by automatch. Default mode uses substring matching, "exact" performs equality matching.
            **unknown : dict
                Keyword arguments not recognized by DataModel. Will issue a warning in the logger.
            """
            self.attr = attr
            self._alias = alias
            self.flags = set(f.strip() for f in flags.split(",")) if flags is not None else set()
            self.args = args or ()
            self.kwargs = kwargs
            self.mask = mask
            self.target = target
            self.flat = flat
            self.dispname = dispname
            self.func = func
            self.qopt = qopt
            self.link = link
            self.filter = filter
            self.proxy = proxy
            self.arg_tf = arg_tf
            self.match = match
            if len(unknown):
                logger.warn("Unknown DataModel parameters: "+", ".join(unknown.keys()))

        def __repr__(self):
            """Return string representation."""
            return "DataModel.Prop({}, {}, {}, {}, {}, {}, {})".format(self.attr, self.alias, ", ".join(self.flags), self.args,
                                                                       self.kwargs, self.mask, self.target, self.dispname,
                                                                       self.flat, self.func, self.link, self.filter, self.qopt)

        @property
        def key(self):
            """Return front facing key.

            Returns
            -------
            str
                alias, if set, attr otherwise.
            """
            return self.alias or self.attr

        def value(self, base, transform="all"):
            """Return value.

            Parameters
            ----------
            base : Insatance or Class
                Base Object
            transform : str, optional
                Transformation to apply. Valid values are:
                    - 'unmask': Return unmasked value
                    - 'raw': Do not apply any transformation.
                    - 'targetval' (Instance only): Get attribute specified by target attribute (unmask if not set)
                    - 'all' (Instance only): Perform dereferencing and function transormations
                    The default is "all".
            """
            if self.proxy is not None:
                foreign = getattr(base, self.proxy)
                foreign._init()
                return foreign._meta.lookup[self.target].value(foreign, transform)
            if transform == "unmask" or (transform == "targetval" and self.target is None):
                return getattr(base, self.mask or self.attr)
            val = getattr(base, self.attr)
            if transform == "raw":
                return val
            elif transform == "targetval":
                return getattr(self.value(base, "raw"), self.target, None)
            if "ref" in self.flags:
                val = {k: self.deref(v) for k, v in val.items()} if isinstance(val, dict) else\
                      [self.deref(v) for v in val] if _isCollection(val) else self.deref(val)
            elif self.func is not None:
                val = self.func(val, *self.args, **self.kwargs)
            elif "call" in self.flags:
                val = base.val(*self.args, **self.kwargs)
            return val

        def resolve(self, Model, query, unmask=False):
            """Resolve foreign columns and add join statements.

            Parameters
            ----------
            Model : Class
                Class deriving from DataModel
            query : SQLAlchemy BaseQuery
                Current query to add joins to
            unmask : boolean, optional
                Unmask final column. The default is False.

            Returns
            -------
            Column
                Column that can be used for filter/sort/match expressions. Automatically aliased.
            SQLAlchemy BaseQuery
                Query with applied joins if necessary
            """
            if self.proxy is not None:
                val = getattr(Model, self.proxy)
                FModel = aliased(inspecc(val).property.mapper.entity)
                FModel._init()
                return FModel._meta.lookup[self.target].resolve(FModel, query.outerjoin(FModel, val), unmask)
            val = getattr(Model, self.attr)
            if unmask or self.target is None:
                return self.value(Model, "unmask" if unmask else "raw"), query
            alias = aliased(inspecc(val).property.mapper.entity)
            column = getattr(alias, self.target)
            return column, query.outerjoin(alias, val)

        def deref(self, value):
            """Dereference value."""
            return None if value is None else value.ref() if self.flat is None else value.ref()[self.flat]

        @property
        def alias(self):
            """Get alias, if set. Otherwise return attr."""
            return self._alias or self.attr

        def writable(self, base):
            """Check whether the attribute is writable."""
            return "patch" in self.flags or (self.value(base) is None and "init" in self.flags)

        def tf(self, value):
            """Transform string argument to filter value.

            If arg_tf is set, the transform is provided entirely by that function. if a ValueError is raised, None is returned.
            If no arg_tf is set, return the value if it is a non-empty string, otherwise None.

            Parameters
            ----------
            value : str
                Value to transform

            Returns
            -------
            any
                Transformed value or None if transformation fails or value is empty
            """
            try:
                if self.arg_tf:
                    return self.arg_tf(value)
                else:
                    return None if len(value) == 0 else value
            except ValueError:
                return None

    class Meta:
        """Class storing the metadata.

        Stores all mapped props and keeps shortcuts for filtering and reverse lookup
        """

        def __init__(self, dictmap, sortables, matchables):
            """Initialize metadata.

            Parameters
            ----------
            dictmap : tuple of tuples
                List of levels, each consisting of a list of properties. Properties can be strings, dicts or Prop instances
            sortables : Collection
                List of prop names that are available for sorting
            """
            self.levels = tuple(tuple(entry if type(entry) == DataModel.Prop else
                                      DataModel.Prop(entry) if type(entry) == str else
                                      DataModel.Prop(**entry) for entry in level) for level in dictmap)
            self.lookup = {prop.key: prop for prop in self.props() if prop.key is not None}
            for s in sortables:
                self.lookup[s].flags.add("sort")
            for m in matchables:
                self.lookup[m].flags.add("match")
            self.filters = tuple(self.props(predicate=lambda prop: prop.filter is not None))
            self.matchables = tuple(self.props(predicate=lambda prop: "match" in prop.flags))

        def props(self, level=None, predicate=lambda x: True):
            """Return list of props available at level, fulfilling the predicate.

            Parameters
            ----------
            level : int, optional
                If set, return only props that are on this level or below. The default is None.
            predicate : function, optional
                Include only props for which predicate returns True. The default is lambda x: True.

            Returns
            -------
            Generator Object
                Iterator listing all Prop objects matching the constraints
            """
            return (prop for lev in range(len(self.levels))
                    for prop in self.levels[lev]
                    if (level is None or lev <= int(level)) and predicate(prop))

    _meta = None

    def __init__(self, props, *args, **kwargs):
        """Default-construct object.

        All arguments are passed on to fromdict() to perform model driven initialization.
        Parameters
        ----------
        props : dict
            Dict containing properties.
        *args : Collection
            Further arguments
        **kwargs : dict
            Further keyword arguments
        """
        self.fromdict(props, *args, **kwargs)

    @classmethod
    def _init(cls):
        """Initialize metadata object."""
        if cls._meta is None:
            cls._meta = cls.Meta(getattr(cls, "_dictmapping_", ((),)),
                                 getattr(cls, "_sortables_", set()),
                                 getattr(cls, "_matchables_", tuple()))

    def ref(self, **kwargs):
        """Generate level 0 representation.

        See todict for more information.
        """
        return self.todict(0, **kwargs)

    def overview(self, **kwargs):
        """Generate level 1 representation.

        See todict for more information.
        """
        return self.todict(1, **kwargs)

    def fulldesc(self, **kwargs):
        """Generate level 2 representation.

        See todict for more information.
        """
        return self.todict(2, **kwargs)

    def todict(self, spec, exclude=set()):
        """Create dictionary representation of the object.

        Parameters
        ----------
        verbosity : int
            Level of detail.
        exclude : set
            Attributes to exclude

        Returns
        -------
        dict
            Dictionary representation
        """
        self._init()
        if isinstance(spec, int):
            propsel = lambda prop: "hidden" not in prop.flags and prop.attr not in exclude and prop.proxy is None
        else:
            sspec = set(spec)
            propsel = lambda prop: prop.attr in sspec and prop.attr not in exclude and prop.proxy is None
            spec = None
        return {prop.key: prop.value(self) for prop in self._meta.props(spec, propsel)}

    @classmethod
    def optimize_query(cls, query, spec):
        """Optimize query by eager loading relationships.

        Parameters
        ----------
        cls : Class
            CLass inheriting from DataModel
        query : SQLAlchemy Query
            Query to add eager loading options to
        spec : int or iterable
            Level of detail or list of attribute names to eager load

        Returns
        -------
        SQLAlchemy Query
            Query with eager loading options added

        """
        cls._init()
        if isinstance(spec, int):
            propsel = lambda prop: "ref" in prop.flags
        else:
            sspec = set(spec)
            propsel = lambda prop: "ref" in prop.flags and prop.attr in sspec
            spec = None
        return query.options(prop.qopt(prop.value(cls, "raw")) for prop in cls._meta.props(spec, propsel))

    @classmethod
    def optimized_query(cls, spec):
        """Generate an optimized query.

        See `optimize_query` for more information.
        """
        return cls.optimize_query(cls.query, spec)

    def fromdict(self, patches, *args, **kwargs):
        """Update object from dictionary representation.

        Parameters
        ----------
        patches : dict
            Dictionary containing updates.
        init : boolean, optional
            Whether the object is initialized (otherwise it is an update operation). The default is False.

        Raises
        ------
        InvalidAttributeError
            An attribute in the patch is not recognized
        MismatchROError
            An attribute that is not writable differs in the patch.

        Returns
        -------
        Instance
            The updated object (self)
        """
        self._init()
        reverse = self._meta.lookup
        for key, value in patches.items():
            if key not in reverse or reverse[key].proxy is not None:
                raise InvalidAttributeError("Unknown attribute '{}'".format(key))
            prop = reverse[key]
            if not prop.writable(self) and prop.value(self) != patches[key]:
                raise MismatchROError("Attribute '{}' is read only and does not match".format(key))
        for key, value in patches.items():
            if key not in reverse:
                continue
            prop = reverse[key]
            if not prop.writable(self):
                continue
            if "ref" not in prop.flags:
                setattr(self, prop.attr, value)
            else:
                attr = getattr(self, prop.attr)
                try:
                    Element = inspecc(getattr(type(self), prop.attr)).property.mapper.entity
                except Exception:
                    logger.warn("Failed to inspect attribute '{}' - ignored".format(prop.attr))
                    continue
                if prop.mask is not None and "managed" not in prop.flags:
                    setattr(self, prop.mask, value)
                elif isinstance(attr, dict) and isinstance(value, dict):
                    elements = {k: Element({prop.link: k, **(value[k] if prop.flat is None else {prop.flat: value[k]})},
                                           self, *args, **kwargs)
                                for k, v in value.items() if k not in attr}  # New elements
                    elements.update({k: a.fromdict(value[k] if prop.flat is None else {prop.flat: value[k]})
                                     for k, a in attr.items() if k in value})  # Updated elements
                    setattr(self, prop.attr, elements)
                elif _isCollection(attr) and _isCollection(value):
                    if "managed" in prop.flags:
                        current = {getattr(val, prop.link) for val in attr}
                        fID = prop.link
                        patch = {val[fID]: val for val in value if fID in val and val[fID] in current} if prop.flat is None else\
                                {val: val for val in value if val in current}
                        new = (val for val in value if fID not in val or val[fID] not in current) if prop.flat is None else\
                              (val for val in value if val not in current)
                        setattr(self, prop.attr,
                                [val.fromdict(patch[getattr(val, fID)], *args, **kwargs)
                                 for val in attr if getattr(val, fID) in patch] +
                                [Element(val, self, *args, **kwargs) for val in new])
                    else:
                        current = {getattr(val, prop.link) for val in attr}
                        requested = set(value)
                        common, new = requested & current, requested - current
                        setattr(self, prop.attr, [val for val in attr if getattr(val, prop.link) in common] +
                                                 [Element(newId, self) for newId in new])
                elif "managed" in prop.flags:
                    if value is None:
                        setattr(self, prop.attr, None)
                    elif attr is None:
                        setattr(self, prop.attr, Element(value, self, *args, **kwargs))
                    else:
                        attr.fromdict(value, *args, **kwargs)
                else:
                    try:
                        attr.fromdict(value, *args, **kwargs)
                    except InvalidAttributeError as err:
                        raise InvalidAttributeError("Cannot patch attribute '{}': {}".format(key, err.args[0])) from None
                    except MismatchROError as err:
                        raise MismatchROError("Cannot patch attribute '{}': {}".format(key, err.args[0])) from None
        return self

    @staticmethod
    def checkCreateParams(data):
        """Validate parameters for creation.

        If parameter validation fails, a descriptive error message should be returned.

        The data object may be modified to allow preprocessing of data.

        Parameters
        ----------
        data : dict
            Dictionary containing the data to create a new object from
        """
        pass

    @classmethod
    def autofilter(cls, query, args):
        """Apply valid filters to the query.

        Valid filters are determined by the `_filters_` class property. Every key in `args` that is a valid filter is added
        to the query.

        Parameters
        ----------
        cls : Class
            Class inheriting from DataModel
        query : Query
            SQLAlchemy Query
        args : dict
            Dictionary containing filters

        Returns
        -------
        Query
            Query with applied filters
        """
        cls._init()
        activeFilters = ((prop, tuple(prop.tf(v) for v in args[prop.key].split(",")))
                         for prop in cls._meta.filters if prop.key in args)
        for prop, values in activeFilters:
            if prop.proxy is not None:
                attr, query = prop.resolve(cls, query, True)
            else:
                attr = prop.value(cls, transform="unmask")
            if prop.filter == "range":
                query = query.filter(attr == values[0]) if len(values) == 1 else\
                        query.filter((attr >= values[0]) & (attr <= values[1]))
            elif prop.filter == "set":
                query = query.filter(or_(attr == v for v in values))
        return query

    @classmethod
    def autosort(cls, query, sorts):
        """Apply valid sort expressions to query.

        Valid sorts are determined by the `_sortables_` class property. Uses values stored in the "sort" key from args.
        Values can be given in the format of "column" or "column,order", where order can be one of "asc" or "desc". If no
        order is specified, "asc" is used by default.
        If more than one sort value is given, all values are applied to the query in order given.

        Parameters
        ----------
        cls : Class
            Class inheriting from DataModel
        query : Query
            SQLAlchemy Query
        sorts : list
            List of sort expressions

        Returns
        -------
        Query
            Query with applied order by expressions
        """
        cls._init()
        for s in sorts:
            column, order = s.split(",", 1) if "," in s else (s, "asc")
            prop = cls._meta.lookup.get(column)
            if prop is None or "sort" not in prop.flags:
                continue
            if prop.target is None:
                column = prop.value(cls, "unmask")
                query = query.order_by(column.desc() if order == "desc" else column.asc())
            else:
                column, query = prop.resolve(cls, query)
                query = query.order_by(func.isnull(column), column.desc() if order == "desc" else column.asc())
        return query

    @classmethod
    def automatch(cls, query, expr, fields=None):
        """Add fuzzy matching to query."""
        cls._init()
        isUnicode = any(ord(c) > 127 for c in expr)
        matchexpr = tuple("%"+substr+"%" for substr in expr.split())
        matchables = cls._meta.matchables if fields is None else (m for m in cls._meta.matchables if m.alias in fields)
        targets = []
        for prop in matchables:
            column, query = prop.resolve(cls, query)
            if not (isUnicode and isinstance(column.type, String) and column.type.charset == "ascii"):
                targets.append((prop, column))
        filters = [column.ilike(match) for match in matchexpr for prop, column in targets if prop.match == "default"] +\
                  [column == prop.tf(expr) for prop, column in targets if prop.match == "exact" and prop.tf(expr) is not None]
        query = query.filter(or_(filter for filter in filters) if filters else False)
        return query.reset_joinpoint()

    def matchvalues(self, fields=None):
        """Return iterator for values relevant for matching."""
        matchables = self._meta.matchables if fields is None else (m for m in self._meta.matchables if m.alias in fields)
        return (prop.value(self, "targetval") for prop in matchables)

    @classmethod
    def augment(cls, props: dict, kwargs: dict):
        """Add keyword arguments to prop dictionary.

        Only arguments that can be resolved to attributes are inserted.

        Parameters
        ----------
        props : dict
            Property dictionary to update
        kwargs : dict
            Keyword arguments dictionary
        """
        cls._init()
        props.update((key, value) for key, value in kwargs.items() if key in cls._meta.lookup)


def _addFlags(kwargs, flags):
    """Add flags to dictmapping entry."""
    kwargs["flags"] = ",".join((kwargs["flags"], flags)) if "flags" in kwargs else flags


def RefProp(attr, mask=None, target=None, **kwargs):
    """Create a reference property."""
    _addFlags(kwargs, "ref")
    return DataModel.Prop(attr, mask=mask, target=target, **kwargs)


def Id(name="ID", **kwargs):
    """Create an ID property."""
    _addFlags(kwargs, "init,sort")
    return DataModel.Prop(name, filter="set", **kwargs)


def Text(name, match=True, **kwargs):
    """Create a name property."""
    _addFlags(kwargs, "sort,match" if match else "sort")
    return DataModel.Prop(name, filter="range", **kwargs)


def Int(name, **kwargs):
    """Create a name property."""
    _addFlags(kwargs, "sort")
    return DataModel.Prop(name, filter=kwargs.pop("filter", "range"), **kwargs)


def Bool(attr, **kwargs):
    """Create a bool property."""
    _addFlags(kwargs, "sort")
    return DataModel.Prop(attr, func=bool, filter="set", arg_tf=_str2bool, **kwargs)


def BoolP(attr, **kwargs):
    """Create a bool property."""
    _addFlags(kwargs, "sort")
    return DataModel.Prop(attr, arg_tf=_str2bool, filter="set", **kwargs)


def Proxy(attr, proxy, **kwargs):
    """Create a proxy property."""
    return DataModel.Prop(attr, target=attr, proxy=proxy, **kwargs)


def Date(attr, time=False, **kwargs):
    """Create a date attribute."""
    format = "%Y-%m-%d %H:%M:%S" if time else "%Y-%m-%d"
    _addFlags(kwargs, "sort")
    return DataModel.Prop(attr, func=lambda date: date.strftime(format) if date else None, filter="range", **kwargs)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

import api
from api.core import API, secure
from api.security import checkPermissions

from datetime import datetime
from flask import jsonify, request
import psutil

from tools.config import Config
from tools.logs import LogReader
from tools.permissions import SystemAdminROPermission


@API.route(api.BaseRoute+"/system/logs", methods=["GET"])
@secure()
def getLogs():
    checkPermissions(SystemAdminROPermission())
    return jsonify(data=sorted(Config["logs"]))


@API.route(api.BaseRoute+"/system/logs/<file>", methods=["GET"])
@secure()
def getLog(file):
    checkPermissions(SystemAdminROPermission())
    log = Config["logs"].get(file)
    if log is None:
        return jsonify(message="Log file not found"), 404
    n = int(request.args.get("n", 10))
    skip = int(request.args.get("skip", 0))
    after = datetime.strptime(request.args["after"], "%Y-%m-%d %H:%M:%S.%f") if "after" in request.args else None
    return jsonify(data=LogReader.tail(log.get("format", "journald"), log["source"], n, skip, after))


@API.route(api.BaseRoute+"/system/updateLog/<int:pid>", methods=["GET"])
@secure()
def getUpdateLog(pid):
    checkPermissions(SystemAdminROPermission())
    log = []
    try:
        with open(Config["options"]["updateLogPath"], 'r') as file:
            log = [line.rstrip() for line in file]
    except Exception as err:
        log = ["Failed to check for updates."]
        API.logger.error("Failed to get update log:"+type(err).__name__+": "+" - ".join(str(arg) for arg in err.args))
    try:
        # Process exists
        proc = psutil.Process(pid)
        processRunning = proc.status() != psutil.STATUS_ZOMBIE
        if not processRunning:
            proc.wait()  # Retrieve process result to prevent zombie apocalypse
        return jsonify({"data": log, "processRunning": processRunning})
    except Exception:
        # Process doesn't exist
        return jsonify({"data": log, "processRunning": False})



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, InvalidUseError
from argparse import ArgumentParser

_statusMap = {0: "active", 1: "suspended", 3: "deleted"}
_statusColor = {0: "green", 1: "yellow", 3: "red"}
_domainAttributes = ("ID", "activeUsers", "address", "adminName", "chat", "chatID", "displayname", "domainStatus",
                     "domainname", "endDay", "homedir", "homeserverID", "inactiveUsers", "maxUser", "orgID", "tel", "title",
                     "virtualUsers")


def _domainStatus(cli, status):
    return cli.col(_statusMap.get(status, "unknown"), _statusColor.get(status, "magenta"))


def _domainQuery(args):
    from .common import domainCandidates
    from orm.domains import Domains
    query = domainCandidates(args.domainspec) if "domainspec" in args else Domains.query
    if "filter" in args and args.filter is not None:
        query = Domains.autofilter(query, {f.split("=", 1)[0]: f.split("=", 1)[1] for f in args.filter if "=" in f})
    if "sort" in args and args.sort is not None:
        query = Domains.autosort(query, args.sort)
    return query


def _dumpDomain(cli, domain):
    displayname = domain.displayname if domain.displayname != domain.domainname else None
    homeserver = cli.col("(local)", attrs=["dark"]) if domain.homeserver is None else \
        "{} ({})".format(domain.homeserver.ID, domain.homeserver.hostname)
    cli.print(cli.col("{} ({}):".format(domain.domainname, domain.ID), attrs=["bold"]))
    cli.print("  ID: "+str(domain.ID))
    cli.print("  orgID: "+str(domain.orgID))
    cli.print("  domainname: "+domain.domainname+(" ({})".format(cli.col(displayname, attrs=["dark"])) if displayname else ""))
    cli.print("  domainStatus: {} ({})".format(domain.domainStatus, _domainStatus(cli, domain.domainStatus)))
    cli.print("  activeUsers: "+str(domain.activeUsers))
    cli.print("  inactiveUsers: "+str(domain.inactiveUsers))
    cli.print("  virtualUsers: "+str(domain.virtualUsers))
    cli.print("  maxUser: "+str(domain.maxUser))
    cli.print("  homeserver: "+homeserver)
    cli.print("  homedir: "+domain.homedir)
    cli.print("  chatID: "+(domain.chatID or cli.col("(none)", attrs=["dark"])) +
              (" ("+cli.col("inactive", "red")+")" if domain.chatID and not domain.chat else ""))
    cli.print("  endDay: "+str(domain.endDay))
    cli.print("  title: "+domain.title)
    cli.print("  address: "+domain.address)
    cli.print("  adminName: "+domain.adminName)
    cli.print("  tel: "+domain.tel)


def _sanitizeData(data):
    cliargs = {"_handle", "_cli", "domainspec", "no_defaults"}
    return {key: value for key, value in data.items() if value is not None and key not in cliargs}


def cliDomainList(args):
    cli = args._cli
    cli.require("DB")
    domains = _domainQuery(args).all()
    if len(domains) == 0:
        cli.print(cli.col("No domains found.", "yellow"))
        return 1
    for domain in domains:
        dname = " ({})".format(cli.col(domain.displayname, attrs=["dark"])) if domain.displayname != domain.domainname else ""
        cli.print("{}: {}{} ({})".format(domain.ID, domain.domainname, dname, _domainStatus(cli, domain.domainStatus)))


def cliDomainShow(args):
    cli = args._cli
    cli.require("DB")
    domains = _domainQuery(args).all()
    if len(domains) == 0:
        cli.print(cli.col("No domains found.", "yellow"))
        return 1
    for domain in domains:
        _dumpDomain(cli, domain)


def cliDomainCreate(args):
    cli = args._cli
    cli.require("DB")
    from orm.domains import Domains
    from orm.misc import DBConf
    from tools.misc import RecursiveDict
    if args.no_defaults:
        data = {}
    else:
        data = DBConf.getFile("grommunio-admin", "defaults-system", True).get("domain", RecursiveDict())
    data.update(RecursiveDict(_sanitizeData(args.__dict__)))
    result, code = Domains.create(data, createRole=data.pop("create_role", False))
    if code != 201:
        cli.print(cli.col("Could not create domain: "+result, "red"))
        return 1
    _dumpDomain(cli, result)


def cliDomainDeleteRecover(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from .common import domainCandidates
    domains = domainCandidates(args.domainspec).all()
    if len(domains) == 0:
        cli.print(cli.col("No domains found.", "yellow"))
        return 1
    if len(domains) > 1:
        cli.print(cli.col("'{}' is ambiguous".format(args.domainspec), "yellow"))
        return 2
    domain = domains[0]
    if args.delete:
        domain.delete()
    else:
        domain.recover()
    DB.session.commit()
    _dumpDomain(cli, domain)


def cliDomainPurge(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from .common import domainCandidates
    domains = domainCandidates(args.domainspec).all()
    if len(domains) == 0:
        cli.print(cli.col("No domains found.", "yellow"))
        return 1
    if len(domains) > 1:
        cli.print(cli.col("'{}' is ambiguous".format(args.domainspec), "yellow"))
        return 2
    domain = domains[0]
    if not args.yes:
        if cli.confirm("Permanently delete domain " +
                       cli.col(domain.domainname, "red", attrs=["bold"]) +
                       (" and all associated files" if args.files else "")+"? [y/N]: "):
            return 1
    domain.purge(deleteFiles=args.files, printStatus=True)
    cli.print("Removing database entries...", end="")
    DB.session.commit()
    cli.print("Done\nDomain removed.")


def cliDomainModify(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from .common import domainCandidates
    domains = domainCandidates(args.domainspec).all()
    if len(domains) == 0:
        cli.print(cli.col("No domains found.", "yellow"))
        return 1
    if len(domains) > 1:
        cli.print(cli.col("'{}' is ambiguous".format(args.domainspec), "yellow"))
        return 2
    domain = domains[0]
    data = _sanitizeData(args.__dict__)
    try:
        domain.fromdict(data)
        DB.session.commit()
    except ValueError as err:
        cli.print(cli.col("Cannot update domain: "+err.args[0]))
        DB.session.rollback()
    _dumpDomain(cli, domain)


def cliDomainQuery(args):
    cli = args._cli
    cli.require("DB")
    attrTf = {"domainStatus": lambda v: cli.col(str(v)+"/", attrs=["dark"])+_domainStatus(cli, v)}\
        if args.format == "pretty" else {}

    from .common import Table
    from orm.domains import Domains
    args.attributes = args.attributes or ("ID", "domainname", "domainStatus")
    query = _domainQuery(args)
    query = Domains.optimize_query(query, args.attributes)
    domains = [domain.todict(args.attributes) for domain in query]
    separator = args.separator or ("," if args.format == "csv" else "  ")
    data = [[attrTf.get(attr, lambda x: x)(domain.get(attr)) for attr in args.attributes] for domain in domains]
    header = None if len(args.attributes) <= 1 and len(data) <= 1 and args.format == "pretty" else args.attributes
    table = Table(data, header, separator, cli.col("(no results)", attrs=["dark"]))
    table.dump(cli, args.format)


def _cliDomainDomainspecAutocomp(prefix, **kwarg):
    from .common import domainCandidates
    from orm.domains import Domains
    return (domain.domainname for domain in domainCandidates(prefix).with_entities(Domains.domainname).all())


def _noComp(**kwargs):
    return ()


def _setupCliDomain(subp: ArgumentParser):
    class AttrChoice:
        def __contains__(self, value):
            return value == [] or value in _domainAttributes

        def __getitem__(self, i):
            return _domainAttributes[i]

        def __len__(self):
            return len(_domainAttributes)

    def addProperties(parser, init):
        parser.add_argument("-u", "--maxUser", required=init, type=int, help="Maximum number of users")
        parser.add_argument("--address", help="Domain contact address")
        parser.add_argument("--adminName", help="Name of the domain admin")
        parser.add_argument("--endDay", help="Domain expiry date in YYYY-MM-DD format")
        parser.add_argument("--orgID", type=int, help="ID of the organization")
        parser.add_argument("--tel", help="Domain contact telephone number")
        parser.add_argument("--title", help="Domain title")

    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    create = sub.add_parser("create", help="Create new domain")
    create.set_defaults(_handle=cliDomainCreate)
    create.add_argument("domainname", help="Name of the domain")
    create.add_argument("--create-role", action="store_true", help="Create domain administrator role for new domain")
    create.add_argument("--homeserver", type=int, help="SID of the server to create the domain on")
    create.add_argument("--no-defaults", action="store_true", help="Do not apply configured default values")
    addProperties(create, True)
    delete = sub.add_parser("delete", help="Soft delete domain",
                            description="Set domain status to deleted and deactivate users")
    delete.set_defaults(_handle=cliDomainDeleteRecover, delete=True)
    delete.add_argument("domainspec", help="Domain ID or prefix to match domainname against")\
        .completer = _cliDomainDomainspecAutocomp
    list = sub.add_parser("list", help="List domains")
    list.set_defaults(_handle=cliDomainList)
    list.add_argument("domainspec", nargs="?", help="Domain ID or prefix to match domainname against")
    list.add_argument("-f", "--filter", nargs="*", help="Filter by attribute, e.g. -f ID=42")
    list.add_argument("-s", "--sort", nargs="*", help="Sort by attribute, e.g. -s domainname,desc")
    modify = sub.add_parser("modify", help="Modify domain")
    modify.set_defaults(_handle=cliDomainModify)
    modify.add_argument("domainspec", help="Domain ID or prefix to match domainname against")\
        .completer = _cliDomainDomainspecAutocomp
    addProperties(modify, False)
    purge = sub.add_parser("purge", help="Permanently delete domain")
    purge.set_defaults(_handle=cliDomainPurge)
    purge.add_argument("domainspec", nargs="?", help="Domain ID or prefix to match domainname against").completer = _noComp
    purge.add_argument("-f", "--files", action="store_true", help="Delete domain and user files on disk")
    purge.add_argument("-y", "--yes", action="store_true", help="Do not question the elevated one")
    query = sub.add_parser("query", help="Query specific domain attributes")
    query.set_defaults(_handle=cliDomainQuery)
    query.add_argument("-f", "--filter", action="append", help="Filter by attribute, e.g. -f ID=42")
    query.add_argument("--format", choices=("csv", "json-flat", "json-structured", "pretty"), help="Set output format",
                       metavar="FORMAT", default="pretty")
    query.add_argument("--separator", help="Set column separator")
    query.add_argument("-s", "--sort", action="append", help="Sort by attribute, e.g. -s username,desc")
    query.add_argument("attributes", nargs="*", choices=AttrChoice(), help="Attributes to query", metavar="ATTRIBUTE")
    recover = sub.add_parser("recover", help="Recover soft-deleted domain")
    recover.set_defaults(_handle=cliDomainDeleteRecover, delete=False)
    recover.add_argument("domainspec", help="Domain ID or prefix to match domainname against")\
        .completer = _cliDomainDomainspecAutocomp
    show = sub.add_parser("show", help="Show detailed information about one or more domains")
    show.set_defaults(_handle=cliDomainShow)
    show.add_argument("domainspec", help="Domain ID or name").completer = _cliDomainDomainspecAutocomp
    show.add_argument("-f", "--filter", nargs="*", help="Filter by attribute, e.g. -f ID=42")
    show.add_argument("-s", "--sort", nargs="*", help="Sort by attribute, e.g. -s domainname,desc")


@Cli.command("domain", _setupCliDomain, help="Domain management")
def cliDomainStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 grommunio GmbH

import logging

from orm import DB
from services import Service, ServiceUnavailableError
from sqlalchemy.exc import IntegrityError
from tools.DataModel import MismatchROError, InvalidAttributeError
from tools.misc import RecursiveDict

logger = logging.getLogger("ldap")


def downsyncUser(user, externID=None):
    """Synchronize a user from LDAP.

    Must not be called on group objects.

    Parameters
    ----------
    user : orm.Users
        User object to synchronize
    externID : bytes, optional
        Reassign externID to this ID. The default is None.

    Returns
    -------
    str
        Message
    int
        HTTP-like result code
    """
    userdata = None
    with Service("ldap", user.orgID, errors=Service.SUPPRESS_INOP) as ldap:
        userdata = ldap.downsyncUser(externID or user.externID, user.properties)

    if userdata is None:
        return "Failed to get user data", 500
    try:
        try:
            user.fromdict(userdata)
        except ServiceUnavailableError:
            logger.warning(f"Failed to synchronize store of user {user.username} - service unavailable")
        user.externID = externID or user.externID
        DB.session.commit()
        return "success", 200
    except (InvalidAttributeError, MismatchROError, ValueError) as err:
        DB.session.rollback()
        return err.args[0], 400
    except IntegrityError as err:
        DB.session.rollback()
        return err.orig.args[1], 400


def downsyncGroup(mlist, externID=None):
    """Synchronize group from LDAP.

    Must not be called on user object.

    Parameters
    ----------
    mlist : orm.MLists
        Group object to synchronize
    externID : bytes, optional
        Reassign externID to this ID. The default is None.

    Returns
    -------
    str
        Message
    int
        HTTP-like result code
    """
    listdata = None
    with Service("ldap", mlist.user.orgID, errors=Service.SUPPRESS_INOP) as ldap:
        listdata = ldap.downsyncUser(externID or mlist.user.externID, mlist.user.properties)

    if listdata is None:
        return "Failed to get group data", 500
    try:
        mlist.fromdict(listdata)
        mlist.user.externID = externID or mlist.user.externID
        DB.session.commit()
        return "success", 200
    except (InvalidAttributeError, MismatchROError, ValueError) as err:
        DB.session.rollback()
        return err.args[0], 400
    except IntegrityError as err:
        DB.session.rollback()
        return err.orig.args[1], 400


def downsyncObject(user, externID=None, syncGroupMembers=False):
    """Synchronize user object from LDAP.

    Dispatches to appropriate synchronization function (user or group)
    and updates group members if requested.

    Parameters
    ----------
    user : TYPE
        DESCRIPTION.
    externID : TYPE, optional
        DESCRIPTION. The default is None.
    syncGroupMembers : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    str
        Message
    int
        HTTP-like status code

    """
    if user.properties["displaytypeex"] != 1:
        return downsyncUser(user, externID)
    from orm.mlists import MLists
    mlist = MLists.query.filter(MLists.listname == user.username).first()
    if not mlist:
        return "No such group", 400
    message, code = downsyncGroup(mlist, externID)
    if syncGroupMembers:
        from services import Service
        with Service("ldap", mlist.user.orgID) as ldap:
            syncGroupMembers(mlist.user.orgID, ldap.getUserInfo(mlist.user.externID))
    return message, code


def importContact(candidate, ldap, orgID, syncExisting=False, domains=None, **kwargs):
    """Import contact from LDAP.

    Parameters
    ----------
    candidate : services.ldap.SearchResult
        Object to synchronize
    ldap : services.ldap.LdapService
        LDAP connection to use
    orgID : int
        Organization ID to limit import to
    syncExisting : bool, optional
        Whether to implicitely update existing objects. The default is False.
    domains : Iterable[int], optional
        Set of domain IDs to import contacts to or None for no restriction. The default is None

    Returns
    -------
    synced : list[orm.users.Users]
        List of synchronized users
    imported : list[orm.users.Users]
        List of imported users
    errors : list[tuple[str, int]]
        List of (message, code) tuple containing error information.
    """
    from orm.domains import Domains
    from orm.users import Users

    synced = []
    imported = []
    errors = []
    domains = Domains.query.filter(Domains.orgID == orgID, Domains.ID.in_(domains) if domains else True)\
                           .with_entities(Domains.ID, Domains.domainname).all()
    existing = Users.query.filter(Users.orgID == orgID, Users.externID == candidate.ID).all()
    if syncExisting:
        for user in existing:
            error, code = downsyncUser(user)
            if code != 200:
                errors.append((f"Failed to synchronize {user.username}: {error}", code))
            else:
                synced.append(user)
    existingDomains = {user.domainID for user in existing}
    domains = [domain for domain in domains if domain.ID not in existingDomains]
    for domain in domains:
        contactData = ldap.downsyncUser(candidate.ID)
        contactData["domainID"] = domain.ID
        result, code = Users.mkContact(contactData, candidate.ID)
        if code != 201:
            errors.append((f"Failed to import contact {candidate.email} into {domain.domainname}: {result}", code))
        else:
            imported.append(result)
    return synced, imported, errors


def importUser(candidate, ldap, force=False, lang="", **kwargs):
    """Import user from LDAP

    Parameters
    ----------
    candidate : services.ldap.SearchResult
        Object to synchronize
    ldap : services.ldap.LdapService
        LDAP connection to use
    force : bool, optional
        Force import if user exists and is linked to a different object. The default is False.
    lang : str, optional
        Language to set for the user. The default is "".

    Returns
    -------
    str | orm.Users
        User object if successful, error message otherwise
    int
        HTTP-like result code
    """
    from orm.misc import DBConf
    from orm.domains import Domains
    from orm.users import Users

    existing = Users.query.filter(Users.username == candidate.email).first()
    if existing:
        if existing.externID == candidate.ID or force:
            result, code = existing, downsyncUser(existing, candidate.ID)[1]
            return existing if code == 200 else result, code
        msg = "and is linked to another LDAP object" if existing.externID else "locally"
        return candidate.type.capitalize()+" already exists "+msg, 400

    domain = Domains.query.filter(Domains.domainname == candidate.email.split("@")[1]).with_entities(Domains.ID).first()
    defaults = RecursiveDict({"user": {}, "domain": {}})
    defaults.update(DBConf.getFile("grommunio-admin", "defaults-system", True))
    defaults.update(DBConf.getFile("grommunio-admin", "defaults-domain-"+str(domain.ID)))
    defaults = defaults.get("user", {})

    userdata = ldap.downsyncUser(candidate.ID)
    defaults.update(RecursiveDict(userdata))
    defaults["lang"] = lang or defaults.get("lang", "")
    return Users.create(defaults, externID=candidate.ID)


def importGroup(candidate, ldap, force=False, syncMembers=False, **kwargs):
    """Import group from LDAP.

    Parameters
    ----------
    candidate : services.ldap.SearchResult
        Object to synchronize
    ldap : services.ldap.LdapService
        LDAP connection to use
    force : bool, optional
        Force import if user exists and is linked to a different object. The default is False.
    syncMembers : bool, optional
        Whether to sync group members after importing. The default is False.

    Returns
    -------
    str | orm.MLists
        Group object if successful, error message otherwise
    int
        HTTP-like result code
    """
    from orm.mlists import MLists
    from orm.users import Users
    xuser = Users.query.filter(Users.username == candidate.email).first()
    xlist = MLists.query.filter(MLists.listname == candidate.email).first()
    if xuser and xlist:
        if xuser.externID == candidate.ID or force:
            result, code = downsyncGroup(xlist, candidate.ID)
            return xlist if code == 200 else result, code
        msg = "and is linked to another LDAP object" if xuser.externID else "locally"
        return candidate.type.capitalize()+" already exists "+msg, 400

    listdata = ldap.downsyncUser(candidate.ID)
    error = MLists.checkCreateParams(listdata)
    if error:
        return error, 400
    try:
        mlist = MLists(listdata)
        mlist.user.externID = candidate.ID
        DB.session.add(mlist)
        DB.session.commit()
        if syncMembers:
            syncGroupMembers(mlist.user.orgID, candidate, ldap)
        return mlist, 201
    except Exception as err:
        return type(err).__name__+": "+" - ".join(str(arg) for arg in err.args), 500


def importObject(candidate, ldap, **kwargs):
    """Import LDAP object

    Automatically calls the appropriate function according to the candidates type

    Parameters
    ----------
    candidate : services.ldap.SearchResult
        LDAP object to import
    ldap : services.ldap.LdapService
        LDAP connection to use
    **kwargs : Any
        Keyword arguments forwarded to the importer

    Raises
    ------
    TypeError
        LDAP object has an invalid type

    Returns
    -------
    dict
        Result object
    int
        HTTP-like status code
    """
    if candidate.type == "contact":
        synced, imported, failed = importContact(candidate, ldap, **kwargs)
        return {"message": f"{len(synced)} synced, {len(imported)} imported, {len(failed)} failed"}, 200
    if candidate.type == "group":
        result, code = importGroup(candidate, ldap, **kwargs)
        return {"message": result} if isinstance(result, str) else result.user.fulldesc(), code
    if candidate.type == "user":
        result, code = importUser(candidate, ldap, **kwargs)
        return {"message": result} if isinstance(result, str) else result.fulldesc(), code
    raise TypeError(f"Unknown object type '{candidate.type}'")


def syncGroupMembers(orgID, ldapgroup, ldap, users=None):
    """Synchronize group members.

    Parameters
    ----------
    orgID : int
        Organization ID to limit members to.
    ldapgroup : services.ldap.SearchResult
        LDAP group object to synchronize.
    ldap : services.ldap.LdapService
        LDAP connection to use
    users : set[bytes], optional
        Set containing LDAP object IDs of imported users or None to determine automatically. The default is None.

    Returns
    -------
    int | NoneType
        Number of users that were added to the group or None if group not found
    int | NoneType
        Number of users that were removed from the group or None if group not found
    """
    from orm.mlists import Associations, MLists
    group = MLists.query.filter(MLists.listname == ldapgroup.email).first()
    if group is None or group.user.orgID != orgID:
        return None, None
    if users is None:
        from orm.users import Users
        users = {user.externID for user in Users.query.filter(Users.orgID == orgID, Users.externID != None)}
    assocs = {assoc.username: assoc for assoc in Associations.query.filter(Associations.listID == group.ID).all()}
    add = 0
    for member in ldap.searchUsers(attributes="idonly", customFilter=ldap.groupMemberFilter(ldapgroup.DN)):
        assoc = assocs.pop(member.email, None)
        if assoc or member.ID not in users:  # Do nothing if already associated or not known
            continue
        add += 1
        DB.session.add(Associations(member.email, group.ID))
    remove = len(assocs)
    for assoc in assocs.values():
        DB.session.delete(assoc)
    DB.session.commit()
    return add, remove



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import DB
from .users import Users
from tools.DataModel import DataModel, Id, Int, RefProp, Text
from tools import formats

from sqlalchemy import Column, ForeignKey
from sqlalchemy.dialects.mysql import INTEGER, TINYINT, VARCHAR
from sqlalchemy.orm import relationship, selectinload, validates


class Associations(DataModel, DB.Base):
    __tablename__ = "associations"

    listID = Column("list_id", INTEGER(10), nullable=False, primary_key=True)
    username = Column("username", VARCHAR(128), nullable=False, primary_key=True)

    _dictmapping_ = ((Id("listID", flags="init"), Text("username", flags="init")),)

    def fromdict(self, username, listID=None, *args, **kwargs):
        self.username = username
        if listID is not None:
            self.listID = listID
        return self


class Specifieds(DataModel, DB.Base):
    __tablename__ = "specifieds"

    ID = Column("id", INTEGER(10, unsigned=True), nullable=False, primary_key=True)
    username = Column("username", VARCHAR(128), nullable=False)
    listID = Column("list_id", INTEGER(10), nullable=False, index=True)

    _dictmapping_ = ((Id(),  Text("username", flags="patch")),
                     (Id("listID", flags="patch"),))

    def fromdict(self, username, *args, **kwargs):
        self.username = username
        return self


class MLists(DataModel, DB.Base):
    __tablename__ = "mlists"

    ID = Column("id", INTEGER(10, unsigned=True), nullable=False, primary_key=True)
    listname = Column("listname", VARCHAR(128), ForeignKey(Users.username), nullable=False, unique=True)
    domainID = Column("domain_id", INTEGER(10, unsigned=True), index=True)
    listType = Column("list_type", TINYINT, nullable=False)
    listPrivilege = Column("list_privilege", TINYINT, nullable=False, server_default="0")

    _dictmapping_ = ((Id(), Text("listname", flags="init"), Int("listType", flags="patch", filter="set")),
                     (Id("domainID", flags="init"),
                      Int("listPrivilege", flags="patch", filter="set")),
                     (RefProp("associations", flags="patch, managed", link="username", flat="username", qopt=selectinload),
                      RefProp("specifieds", flags="patch, managed", link="username", flat="username", qopt=selectinload),
                      {"attr": "displayname", "flags": "patch"},
                      Int("hidden", flags="patch"),
                      RefProp("user", flat="ID")))

    user = relationship(Users, primaryjoin=listname == Users.username, foreign_keys=listname, cascade="all, delete-orphan",
                        single_parent=True, back_populates="mlist")
    associations = relationship(Associations, primaryjoin=ID == Associations.listID, foreign_keys=Associations.listID,
                                cascade="all, delete-orphan", single_parent=True)
    specifieds = relationship(Specifieds, primaryjoin=ID == Specifieds.listID, foreign_keys=Specifieds.listID,
                              cascade="all, delete-orphan", single_parent=True)

    TYPE_NORMAL = 0
    TYPE_DOMAIN = 2

    PRIV_ALL = 0
    PRIV_INTERNAL = 1
    PRIV_DOMAIN = 2
    PRIV_SPECIFIED = 3
    PRIV_OUTGOING = 4

    @classmethod
    def checkCreateParams(cls, data):
        from .domains import Domains
        from .users import Users
        if "listname" not in data:
            return "Missing list name"
        if "domainID" in data:
            domain = Domains.query.filter(Domains.ID == data["domainID"]).first()
        elif "@" in data["listname"]:
            domain = Domains.query.filter(Domains.domainname == data["listname"].split("@")[1]).first()
        else:
            domain = None
        if domain is None:
            return "Invalid Domain"
        data["domainID"] = domain.ID
        data["domain"] = domain
        if "@" not in data["listname"]:
            data["listname"] += "@"+domain.domainname
        elif data["listname"].split("@")[1] != domain.domainname:
            return "Domain specifications mismatch"
        if "listType" not in data:
            return "Missing list type"
        if Users.query.filter(Users.username == data["listname"]).count() > 0:
            return "User exists"
        elif data["listType"] not in (cls.TYPE_NORMAL, cls.TYPE_DOMAIN):
            return "Unsupported list type"
        if data.get("listPrivilege", 0) not in (cls.PRIV_ALL,
                                                cls.PRIV_INTERNAL,
                                                cls.PRIV_DOMAIN,
                                                cls.PRIV_SPECIFIED,
                                                cls.PRIV_OUTGOING):
            return "Invalid privilege"

    def __init__(self, props, *args, **kwargs):
        from .users import Users
        self.domain = props.pop("domain")
        self.listType = props.pop("listType", 0)
        self.listPrivilege = props.pop("listPrivilege", 0)
        self.user = Users({"username": props.get("listname"),
                           "domainID": self.domain.ID,
                           "domain": self.domain,
                           "domainStatus": self.domain.domainStatus,
                           "properties": {"displaytypeex": 1}})
        self.user.maildir = ""
        self.fromdict(props, *args, **kwargs)

    @property
    def displayname(self):
        return self.user.properties.get("displayname", "")

    @displayname.setter
    def displayname(self, value):
        self.user.properties["displayname"] = value

    @property
    def hidden(self):
        return self.user.properties.get("attributehidden_gromox", 0)

    @hidden.setter
    def hidden(self, value):
        self.user.properties["attributehidden_gromox"] = value

    @validates("listname")
    def validateListname(self, key, value, *args):
        if not formats.email.match(value):
            raise ValueError(f"{value} is not a valid e-ail address")
        return value

    @validates("associations")
    def validateAssociations(self, key, assoc, *args):
        if self.listType != self.TYPE_NORMAL:
            raise ValueError("Direct user association is only possible for normal mailing lists")
        return assoc

    @validates("specifieds")
    def validateSpecifieds(self, key, spec, *args):
        if self.listPrivilege != self.PRIV_SPECIFIED:
            raise ValueError("Sender specification requires 'specific' list privilege")
        return spec

    @validates("listType")
    def validateListType(self, key, type):
        if type not in range(3):
            raise ValueError("Invalid list type")
        return type

    @validates("listPrivilege")
    def validateListPrivilege(self, key, priv):
        if priv not in range(5):
            raise ValueError("Invalid list privilege")
        return priv



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

import re

email = re.compile(r"^(?P<user>[a-zA-Z0-9_.+-]+)@(?P<domain>[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)$")
domain = re.compile(r"^[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$")



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import ServiceHub, ServiceUnavailableError

#from mattermostdriver import Driver
#from mattermostdriver.exceptions import InvalidOrMissingParameters, ResourceNotFound, ContentTooLarge
from requests.exceptions import ConnectionError, HTTPError

import hashlib
import random
import string


def handleGrochatExceptions(service, error):
    if isinstance(error, ConnectionError):
        return ServiceHub.SUSPENDED
    elif isinstance(error, (InvalidOrMissingParameters, ResourceNotFound, ContentTooLarge)):
        raise ValueError("grochat: invalid data")


@ServiceHub.register("chat", handleGrochatExceptions, maxreloads=3)
class GrochatService:
    def __init__(self):
        from tools.config import Config
        try:
            self.driver = Driver(Config["chat"]["connection"])
            self.driver.login()
        except Exception as err:
            raise ServiceUnavailableError("Failed to connect to grommunio-chat", err)

    @staticmethod
    def _addif(props, data, srcname, dstname):
        if srcname in props:
            data[dstname] = props[srcname]

    @staticmethod
    def userToData(user):
        userdata = {"email": user.username, "username": user.username.replace("@", "_")}
        props = user.properties
        GrochatService._addif(props, userdata, "givenname", "first_name")
        GrochatService._addif(props, userdata, "surname", "last_name")
        GrochatService._addif(props, userdata, "nickname", "nickname")
        if user.chatID:
            userdata["id"] = user.chatID
        return userdata

    def createUser(self, user):
        """Create grochat user from grommunio user."""
        if user.chatID:
            return self.driver.users.get_user(user.chatID)
        userdata = self.userToData(user)
        userdata["auth_service"] = "pam"
        userdata["password"] = "".join(random.choices(string.ascii_letters+string.digits, k=16))
        gcUser = self.driver.users.create_user(userdata)
        user.chatID = gcUser["id"]
        self.linkUser(user)
        return gcUser

    def updateUser(self, user, create=False):
        """Update grochat user from grommunio user."""
        isdata = isinstance(user, dict)
        if not isdata and user.chatID is None:
            return self.createUser(user) if create else None
        userdata = user if isdata else self.userToData(user)
        return self.driver.users.patch_user(userdata["id"], userdata)

    def setUserRoles(self, userID, roles):
        return self.driver.users.update_user_role(userID, {"roles": roles})

    def linkUser(self, user):
        """Add user to domain team."""
        if not user.chatID or not user.domain.chatID:
            return None
        return self.driver.teams.add_user_to_team(user.domain.chatID, {"team_id": user.domain.chatID, "user_id": user.chatID})

    def activateUser(self, user, status):
        """Set user active status.

        Parameters
        ----------
        user : orm.users.Users
            User to (de)activate
        status : bool
            Active status of the user
        """
        if not user.chatID:
            return None
        return self.driver.users.update_user_active_status(user.chatID, {"active": status})

    def deleteUser(self, user):
        """Permanently delete a user.

        Automatically resets the user's chatID.

        Parameters
        ----------
        user : orm.users.Users
            grommunio user object
        """
        if not user.chatID:
            return None
        res = self.driver.client.make_request("delete", "/users/"+user.chatID, params={"permanent": "true"})
        if res.ok:
            user.chatID = None
        return res

    def getUser(self, userID):
        if not userID:
            return None
        try:
            return self.driver.users.get_user(userID)
        except HTTPError:
            return None

    def domainToData(self, domain):
        teamname = hashlib.md5(domain.domainname.encode("ascii")).hexdigest()
        teamdata = {"name": teamname, "display_name": domain.title or domain.domainname, "type": "I"}
        if domain.chatID:
            teamdata["id"] = domain.chatID
        return teamdata

    def createTeam(self, domain):
        """Create team for domain."""
        if domain.chatID:
            return self.driver.teams.get_team(domain.chatID)
        teamdata = self.domainToData(domain)
        gcTeam = self.driver.teams.create_team(teamdata)
        domain.chatID = gcTeam["id"]
        return gcTeam

    def activateTeam(self, domain, status):
        """Archive a domains team."""
        if not domain.chatID:
            return None
        if status:
            resp = self.driver.client.make_request("post", "/teams/"+domain.chatID+"/restore")
            return resp.json() if resp.status_code == 200 else None
        else:
            return self.driver.teams.delete_team(domain.chatID)

    def getTeam(self, teamID):
        if not teamID:
            return None
        return self.driver.teams.get_team(teamID)

    def updateTeam(self, domain):
        if not domain.chatID:
            return None
        teamdata = self.domainToData(domain)
        return self.driver.teams.update_team(domain.chatID, teamdata)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, InvalidUseError
from argparse import ArgumentParser

def _human(size):
    units = ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
    index = 0
    while size >= 1024 and index < len(units)-1:
        index += 1
        size /= 1024
    return size, units[index]


def _du(path):
    import os
    files, size = 0, os.path.getsize(path)
    for pathname, dirnames, filenames in os.walk(path):
        for filename in dirnames+filenames:
            file = os.path.join(pathname, filename)
            if not os.path.islink(file):
                files += os.path.isfile(file)
                size += os.path.getsize(file)
    return files, size


def _statStr(cli, files, size):
    human = "" if size < 1024 else " ("+cli.col("{:.3n} {}".format(*_human(size)), attrs=["bold"])+")"
    return f"{size:,} bytes{human} used by "+cli.col(f"{files} file"+("" if files == 1 else "s"), attrs=["bold"])


def cliFsDu(args):
    cli = args._cli
    from tools.config import Config
    files = size = 0
    if args.partition is None or args.partition == "domain":
        prefix = Config["options"]["domainPrefix"]
        f, s = _du(prefix)
        files += f
        size += s
        cli.print(prefix+": "+_statStr(cli, f, s))
    if args.partition is None or args.partition == "user":
        prefix = Config["options"]["userPrefix"]
        f, s = _du(prefix)
        files += f
        size += s
        cli.print(prefix+": "+_statStr(cli, f, s))
    if args.partition == None:
        cli.print(_statStr(cli, files, size))


def _clean(cli, path, used, maxdepth, du=False, delete=True):
    import os
    import shutil
    path = path.rstrip(os.path.sep)
    prefix = len(path)
    maxdepth -= 1
    files = size = 0
    for pathname, dirnames, filenames in os.walk(path):
        depth = pathname[prefix:].count(os.path.sep)
        if depth > maxdepth:
            dirnames.clear()
            continue
        if depth < maxdepth:
            continue
        removed = 0
        for dirname in dirnames:
            dp = os.path.join(pathname, dirname)
            if dp not in used:
                removed += 1
                if du:
                    f, s = _du(dp)
                    files += f
                    size += s
                cli.print("Remov{} {}".format("ing" if delete else "e", cli.col(dp, attrs=["bold"])))
                if delete:
                    shutil.rmtree(dp, ignore_errors=True)
        if len(dirnames)-removed <= 0 and depth != 0:
            size += os.path.getsize(pathname)
            cli.print("Remov{} empty directory {}".format("ing" if delete else "e", cli.col(pathname, attrs=["bold"])))
            if delete:
                try: os.rmdir(pathname)
                except: pass
        dirnames.clear()
    return files, size


def cliFsClean(args):
    cli = args._cli
    cli.require("DB")
    from tools.config import Config
    opt = Config["options"]
    files = size = 0
    if args.partition is None or args.partition == "domain":
        from orm.domains import Domains
        used = {d.homedir for d in Domains.query.with_entities(Domains.homedir).filter(Domains.homedir != "").all()}
        f, s = _clean(cli, opt["domainPrefix"], used, opt["domainStorageLevels"], not args.nostat, not args.dryrun)
        files += f
        size += s
    if args.partition is None or args.partition == "user":
        from orm.users import Users
        used = {u.maildir for u in Users.query.with_entities(Users.maildir).filter(Users.maildir != "").all()}
        f, s = _clean(cli, opt["userPrefix"], used, opt["userStorageLevels"], not args.nostat, not args.dryrun)
        files += f
        size += s
    if not args.nostat:
        cli.print(("Operation would free " if args.dryrun else "Freed ")+_statStr(cli, files, size))


def _setupCliFsParser(subp: ArgumentParser):
    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    clean = sub.add_parser("clean", help="Remove unused user and domain files")
    clean.description = "Delete orphaned domain and user directories that may be left behind when removing a domain or user "\
                        "without deleteFiles directive"
    clean.set_defaults(_handle=cliFsClean)
    clean.add_argument("partition", nargs="?", choices=("domain", "user"), help="Clean only specified partition")
    clean.add_argument("-d", "--dryrun", action="store_true", help="Do not actually delete anything")
    clean.add_argument("-s", "--nostat", action="store_true", help="Do not collect disk usage of deleted files")
    du = sub.add_parser("du", help="Show disk usage")
    du.set_defaults(_handle=cliFsDu)
    du.add_argument("partition", nargs="?", choices=("domain", "user"), help="Partition to calculate disk usage for")


@Cli.command("fs", _setupCliFsParser, help="Filesystem operations")
def cliFsStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020-2021 grommunio GmbH

SERIAL_ENDIAN = "little"  # Endianness used for binary serialization


class _ReverseLookup:
    @classmethod
    def lookup(cls, value, default=None):
        if not hasattr(cls, "_lookup"):
            from collections.abc import Hashable
            cls._lookup = {getattr(cls, key): key for key in dir(cls) if not key.startswith("_")
                           if isinstance(getattr(cls, key), Hashable)}
        return cls._lookup.get(value, default)

    @classmethod
    def str(cls, value):
        return cls.lookup(value, hex(value))


class PropTypes(_ReverseLookup):
    UNSPECIFIED = 0x0000
    SHORT = 0x0002
    LONG = 0x0003
    FLOAT = 0x0004
    DOUBLE = 0x0005
    CURRENCY = 0x0006
    FLOATINGTIME = 0x0007
    ERROR = 0x000a
    BYTE = 0x000b
    OBJECT = 0x000d
    LONGLONG = 0x0014
    STRING = 0x001e
    WSTRING = 0x001f
    FILETIME = 0x0040
    GUID = 0x0048
    SVREID = 0x00fb
    RESTRICTION = 0x00fd
    RULE = 0x00fe
    BINARY = 0x0102
    SHORT_ARRAY = 0x1002
    LONG_ARRAY = 0x1003
    LONGLONG_ARRAY = 0x1014
    STRING_ARRAY = 0x101e
    WSTRING_ARRAY = 0x101f
    GUID_ARRAY = 0x1048
    BINARY_ARRAY = 0x1102

    intTypes = {BYTE, SHORT, LONG, ERROR, LONGLONG, FILETIME, CURRENCY}
    floatTypes = {FLOAT, DOUBLE, FLOATINGTIME}

    @classmethod
    def lookup(cls, value, default=None):
        return super(PropTypes, cls).lookup(value & 0xFFFF, default)

    @classmethod
    def pyType(cls, value):
        value = value & 0xFFFF
        return int if value in cls.intTypes else float if value in cls.floatTypes else bytes if value == cls.BINARY else str

    @classmethod
    def ismv(cls, value):
        return bool(value & 0x1000)


class ExchangeErrors(_ReverseLookup):
    ecSuccess = 0
    ecUnknownUser = 0x000003EB
    ecServerOOM = 0x000003F0
    ecLoginPerm = 0x000003F2
    ecNotSearchFolder = 0x00000461
    ecNoReceiveFolder = 0x00000463
    ecWrongServer = 0x00000478
    ecBufferTooSmall = 0x0000047D
    ecSearchFolderScopeViolation = 0x00000490
    ecRpcFormat = 0x000004B6
    ecNullObject = 0x000004B9
    ecQuotaExceeded = 0x000004D9
    ecMaxAttachmentExceeded = 0x000004DB
    ecNotExpanded = 0x000004F7
    ecNotCollapsed = 0x000004F8
    ecDstNullObject = 0x00000503
    ecMsgCycle = 0x00000504
    ecTooManyRecips = 0x00000505
    RPC_X_BAD_STUB_DATA = 0x000006F7
    ecRejected = 0x000007EE
    ecWarnWithErrors = 0x00040380
    SYNC_W_CLIENT_CHANGE_NEWER = 0x00040821
    ecError = 0x80004005
    STG_E_ACCESSDENIED = 0x80030005
    StreamSeekError = 0x80030019
    ecNotSupported = 0x80040102
    ecInvalidObject = 0x80040108
    ecObjectModified = 0x80040109
    ecNotFound = 0x8004010F
    ecLoginFailure = 0x80040111
    ecUnableToAbort = 0x80040114
    ecRpcFailed = 0x80040115
    ecTooComplex = 0x80040117
    MAPI_E_UNKNOWN_CPID = 0x8004011E
    MAPI_E_UNKNOWN_LCID = 0x8004011F
    ecTooBig = 0x80040305
    MAPI_E_DECLINE_COPY = 0x80040306
    ecTableTooBig = 0x80040403
    ecInvalidBookmark = 0x80040405
    ecNotInQueue = 0x80040601
    ecDuplicateName = 0x80040604
    ecNotInitialized = 0x80040605
    MAPI_E_FOLDER_CYCLE = 0x8004060B
    EC_EXCEEDED_SIZE = 0x80040610
    ecAmbiguousRecip = 0x80040700
    SYNC_E_IGNORE = 0x80040801
    SYNC_E_CONFLICT = 0x80040802
    SYNC_E_NO_PARENT = 0x80040803
    NotImplemented = 0x80040FFF
    ecAccessDenied = 0x80070005
    ecMAPIOOM = 0x8007000E
    ecInvalidParam = 0x80070057


class ExmdbCodes(_ReverseLookup):
    SUCCESS = 0x00
    ACCESS_DENY = 0x01
    MAX_REACHED = 0x02
    LACK_MEMORY = 0x03
    MISCONFIG_PREFIX = 0x04
    MISCONFIG_MODE = 0x05
    CONNECT_INCOMPLETE = 0x06
    PULL_ERROR = 0x07
    DISPATCH_ERROR = 0x08
    PUSH_ERROR = 0x09


class PropTags(_ReverseLookup):
    """
        These tags are also accessed via getattr() calls and thus effectively
        exposed to AWEB.
    """
    ABPROVIDERID = 0x36150102
    ACCESS = 0x0FF40003
    ACCESSCONTROLLISTDATA = 0x3FE00102
    ACCESSLEVEL = 0x0FF70003
    ACCOUNT = 0x3A00001F
    ACCOUNT_STRING8 = 0x3A00001E
    ADDITIONALRENENTRYIDS = 0x36D81102
    ADDITIONALRENENTRYIDSEX = 0x36D90102
    ADDRESSBOOKAUTHORIZEDSENDERS = 0x8CD8000D
    ADDRESSBOOKCONTAINERID = 0xFFFD0003
    ADDRESSBOOKDELIVERYCONTENTLENGTH = 0x806A0003
    ADDRESSBOOKDISPLAYNAMEPRINTABLE = 0x39FF001F
    ADDRESSBOOKDISPLAYNAMEPRINTABLE_STRING8 = 0x39FF001E
    ADDRESSBOOKDISPLAYTYPEEXTENDED = 0x8C930003
    ADDRESSBOOKDISTRIBUTIONLISTEXTERNALMEMBERCOUNT = 0x8CE30003
    ADDRESSBOOKDISTRIBUTIONLISTMEMBERCOUNT = 0x8CE20003
    ADDRESSBOOKDISTRIBUTIONLISTMEMBERSUBMITACCEPTED = 0x8073000D
    ADDRESSBOOKDISTRIBUTIONLISTMEMBERSUBMITREJECTED = 0x8CDA000D
    ADDRESSBOOKDISTRIBUTIONLISTREJECTMESSAGESFROMDLMEMBERS = 0x8CDB000D
    ADDRESSBOOKENTRYID = 0x663B0102
    ADDRESSBOOKEXTENSIONATTRIBUTE1 = 0x802D001F
    ADDRESSBOOKEXTENSIONATTRIBUTE10 = 0x8036001F
    ADDRESSBOOKEXTENSIONATTRIBUTE10_STRING8 = 0x8036001E
    ADDRESSBOOKEXTENSIONATTRIBUTE11 = 0x8C57001F
    ADDRESSBOOKEXTENSIONATTRIBUTE11_STRING8 = 0x8C57001E
    ADDRESSBOOKEXTENSIONATTRIBUTE12 = 0x8C58001F
    ADDRESSBOOKEXTENSIONATTRIBUTE12_STRING8 = 0x8C58001E
    ADDRESSBOOKEXTENSIONATTRIBUTE13 = 0x8C59001F
    ADDRESSBOOKEXTENSIONATTRIBUTE13_STRING8 = 0x8C59001E
    ADDRESSBOOKEXTENSIONATTRIBUTE14 = 0x8C60001F
    ADDRESSBOOKEXTENSIONATTRIBUTE14_STRING8 = 0x8C60001E
    ADDRESSBOOKEXTENSIONATTRIBUTE15 = 0x8C61001F
    ADDRESSBOOKEXTENSIONATTRIBUTE15_STRING8 = 0x8C61001E
    ADDRESSBOOKEXTENSIONATTRIBUTE1_STRING8 = 0x802D001E
    ADDRESSBOOKEXTENSIONATTRIBUTE2 = 0x802E001F
    ADDRESSBOOKEXTENSIONATTRIBUTE2_STRING8 = 0x802E001E
    ADDRESSBOOKEXTENSIONATTRIBUTE3 = 0x802F001F
    ADDRESSBOOKEXTENSIONATTRIBUTE3_STRING8 = 0x802F001E
    ADDRESSBOOKEXTENSIONATTRIBUTE4 = 0x8030001F
    ADDRESSBOOKEXTENSIONATTRIBUTE4_STRING8 = 0x8030001E
    ADDRESSBOOKEXTENSIONATTRIBUTE5 = 0x8031001F
    ADDRESSBOOKEXTENSIONATTRIBUTE5_STRING8 = 0x8031001E
    ADDRESSBOOKEXTENSIONATTRIBUTE6 = 0x8032001F
    ADDRESSBOOKEXTENSIONATTRIBUTE6_STRING8 = 0x8032001E
    ADDRESSBOOKEXTENSIONATTRIBUTE7 = 0x8033001F
    ADDRESSBOOKEXTENSIONATTRIBUTE7_STRING8 = 0x8033001E
    ADDRESSBOOKEXTENSIONATTRIBUTE8 = 0x8034001F
    ADDRESSBOOKEXTENSIONATTRIBUTE8_STRING8 = 0x8034001E
    ADDRESSBOOKEXTENSIONATTRIBUTE9 = 0x8035001F
    ADDRESSBOOKEXTENSIONATTRIBUTE9_STRING8 = 0x8035001E
    ADDRESSBOOKFOLDERPATHNAME = 0x8004001F
    ADDRESSBOOKFOLDERPATHNAME_STRING8 = 0x8004001E
    ADDRESSBOOKHIERARCHICALCHILDDEPARTMENTS = 0x8C9A000D
    ADDRESSBOOKHIERARCHICALDEPARTMENTMEMBERS = 0x8C97000D
    ADDRESSBOOKHIERARCHICALISHIERARCHICALGROUP = 0x8CDD000B
    ADDRESSBOOKHIERARCHICALPARENTDEPARTMENT = 0x8C99000D
    ADDRESSBOOKHIERARCHICALROOTDEPARTMENT = 0x8C98001E
    ADDRESSBOOKHIERARCHICALSHOWINDEPARTMENTS = 0x8C94000D
    ADDRESSBOOKHOMEMESSAGEDATABASE = 0x8006001F
    ADDRESSBOOKHOMEMESSAGEDATABASE_STRING8 = 0x8006001E
    ADDRESSBOOKISMASTER = 0xFFFB000B
    ADDRESSBOOKISMEMBEROFDISTRIBUTIONLIST = 0x8008001E
    ADDRESSBOOKMANAGEDISTRIBUTIONLIST = 0x6704000D
    ADDRESSBOOKMANAGER = 0x8005000D
    ADDRESSBOOKMANAGERDISTINGUISHEDNAME = 0x8005001F
    ADDRESSBOOKMEMBER = 0x8009000D
    ADDRESSBOOKMESSAGEID = 0x674F0014
    ADDRESSBOOKMODERATIONENABLED = 0x8CB5000B
    ADDRESSBOOKNETWORKADDRESS = 0x8170101F
    ADDRESSBOOKNETWORKADDRESS_STRING8 = 0x8170101E
    ADDRESSBOOKOBJECTDISTINGUISHEDNAME = 0x803C001F
    ADDRESSBOOKOBJECTDISTINGUISHEDNAME_STRING8 = 0x803C001E
    ADDRESSBOOKOBJECTGUID = 0x8C6D0102
    ADDRESSBOOKORGANIZATIONALUNITROOTDISTINGUISHEDNAME = 0x8CA8001F
    ADDRESSBOOKORGANIZATIONALUNITROOTDISTINGUISHEDNAME_STRING8 = 0x8CA8001E
    ADDRESSBOOKORGANIZATIONNAME = 0x8101101f
    ADDRESSBOOKORGANIZATIONALUNITNAME = 0x8102101f
    ADDRESSBOOKOWNER = 0x800C000D
    ADDRESSBOOKOWNERBACKLINK = 0x8024000D
    ADDRESSBOOKPARENTENTRYID = 0xFFFC0102
    ADDRESSBOOKPHONETICCOMPANYNAME = 0x8C91001F
    ADDRESSBOOKPHONETICCOMPANYNAME_STRING8 = 0x8C91001E
    ADDRESSBOOKPHONETICDEPARTMENTNAME = 0x8C90001F
    ADDRESSBOOKPHONETICDEPARTMENTNAME_STRING8 = 0x8C90001E
    ADDRESSBOOKPHONETICDISPLAYNAME = 0x8C92001F
    ADDRESSBOOKPHONETICDISPLAYNAME_STRING8 = 0x8C92001E
    ADDRESSBOOKPHONETICGIVENNAME = 0x8C8E001F
    ADDRESSBOOKPHONETICGIVENNAME_STRING8 = 0x8C8E001E
    ADDRESSBOOKPHONETICSURNAME = 0x8C8F001F
    ADDRESSBOOKPHONETICSURNAME_STRING8 = 0x8C8F001E
    ADDRESSBOOKPROXYADDRESSES = 0x800F101F
    ADDRESSBOOKPROXYADDRESSES_STRING8 = 0x800F101E
    ADDRESSBOOKPUBLICDELEGATES = 0x8015000D
    ADDRESSBOOKREPORTS = 0x800E000D
    ADDRESSBOOKROOMCAPACITY = 0x08070003
    ADDRESSBOOKROOMCONTAINERS = 0x8C96101F
    ADDRESSBOOKROOMDESCRIPTION = 0x0809001F
    ADDRESSBOOKROOMDESCRIPTION_STRING8 = 0x0809001E
    ADDRESSBOOKSENDERHINTTRANSLATIONS = 0x8CAC101F
    ADDRESSBOOKSENIORITYINDEX = 0x8CA00003
    ADDRESSBOOKTARGETADDRESS = 0x8011001F
    ADDRESSBOOKTARGETADDRESS_STRING8 = 0x8011001E
    ADDRESSBOOKUNAUTHORIZEDSENDERS = 0x8CD9000D
    ADDRESSBOOKX509CERTIFICATE = 0x8C6A1102
    ADDRESSTYPE = 0x3002001F
    ADDRESSTYPE_STRING8 = 0x3002001E
    ALTERNATERECIPIENTALLOWED = 0x0002000B
    ANR = 0x360C001F
    ANR_STRING8 = 0x360C001E
    ARCHIVEDATE = 0x301F0040
    ARCHIVEPERIOD = 0x301E0003
    ARCHIVETAG = 0x30180102
    ASSISTANT = 0x3A30001F
    ASSISTANTTELEPHONENUMBER = 0x3A2E001F
    ASSISTANTTELEPHONENUMBER_STRING8 = 0x3A2E001E
    ASSISTANT_STRING8 = 0x3A30001E
    ASSOCIATED = 0x67AA000B
    ATTACHADDITIONALINFORMATION = 0x370F0102
    ATTACHCONTENTBASE = 0x3711001F
    ATTACHCONTENTBASE_STRING8 = 0x3711001E
    ATTACHCONTENTID = 0x3712001F
    ATTACHCONTENTID_STRING8 = 0x3712001E
    ATTACHCONTENTLOCATION = 0x3713001F
    ATTACHCONTENTLOCATION_STRING8 = 0x3713001E
    ATTACHDATABINARY = 0x37010102
    ATTACHDATAOBJECT = 0x3701000D
    ATTACHENCODING = 0x37020102
    ATTACHEXTENSION = 0x3703001F
    ATTACHEXTENSION_STRING8 = 0x3703001E
    ATTACHFILENAME = 0x3704001F
    ATTACHFILENAME_STRING8 = 0x3704001E
    ATTACHFLAGS = 0x37140003
    ATTACHLONGFILENAME = 0x3707001F
    ATTACHLONGFILENAME_STRING8 = 0x3707001E
    ATTACHLONGPATHNAME = 0x370D001F
    ATTACHLONGPATHNAME_STRING8 = 0x370D001E
    ATTACHMETHOD = 0x37050003
    ATTACHMIMETAG = 0x370E001F
    ATTACHMIMETAG_STRING8 = 0x370E001E
    ATTACHNUMBER = 0x0E210003
    ATTACHPATHNAME = 0x3708001F
    ATTACHPATHNAME_STRING8 = 0x3708001E
    ATTACHPAYLOADCLASS = 0x371A001F
    ATTACHPAYLOADCLASS_STRING8 = 0x371A001E
    ATTACHPAYLOADPROVIDERGUIDSTRING = 0x3719001F
    ATTACHPAYLOADPROVIDERGUIDSTRING_STRING8 = 0x3719001E
    ATTACHRENDERING = 0x37090102
    ATTACHSIZE = 0x0E200003
    ATTACHTAG = 0x370A0102
    ATTACHTRANSPORTNAME = 0x370C001F
    ATTACHTRANSPORTNAME_STRING8 = 0x370C001E
    ATTACHMENTCONTACTPHOTO = 0x7FFF000B
    ATTACHMENTFLAGS = 0x7FFD0003
    ATTACHMENTHIDDEN = 0x7FFE000B
    ATTACHMENTLINKID = 0x7FFA0003
    ATTRIBUTEHIDDEN = 0x10F4000B
    ATTRIBUTEHIDDEN_GROMOX = 0x10F40003
    ATTRIBUTESYSTEM = 0x10F5000B
    ATTRIBUTEREADONLY = 0x10F6000B
    AUTOFORWARDCOMMENT = 0x0004001F
    AUTOFORWARDCOMMENT_STRING8 = 0x0004001E
    AUTOFORWARDED = 0x0005000B
    AUTORESPONSESUPPRESS = 0x3FDF0003
    BIRTHDAY = 0x3A420040
    BLOCKSTATUS = 0x10960003
    BODY = 0x1000001F
    BODYCONTENTID = 0x1015001F
    BODYCONTENTID_STRING8 = 0x1015001E
    BODYCONTENTLOCATION = 0x1014001F
    BODYCONTENTLOCATION_STRING8 = 0x1014001E
    BODYHTML = 0x1013001F
    BODYHTML_STRING8 = 0x1013001E
    BODY_STRING8 = 0x1000001E
    BUSINESS2TELEPHONENUMBER = 0x3A1B001F
    BUSINESS2TELEPHONENUMBER_STRING8 = 0x3A1B001E
    BUSINESS2TELEPHONENUMBERS = 0x3A1B101F
    BUSINESSFAXNUMBER = 0x3A24001F
    BUSINESSFAXNUMBER_STRING8 = 0x3A24001E
    BUSINESSHOMEPAGE = 0x3A51001F
    BUSINESSHOMEPAGE_STRING8 = 0x3A51001E
    BUSINESSTELEPHONENUMBER = 0x3A08001F
    BUSINESSTELEPHONENUMBER_STRING8 = 0x3A08001E
    CALLID = 0x6806001F
    CALLID_STRING8 = 0x6806001E
    CALLBACKTELEPHONENUMBER = 0x3A02001F
    CALLBACKTELEPHONENUMBER_STRING8 = 0x3A02001E
    CARTELEPHONENUMBER = 0x3A1E001F
    CARTELEPHONENUMBER_STRING8 = 0x3A1E001E
    CDORECURRENCEID = 0x10C50040
    CHANGEKEY = 0x65E20102
    CHANGENUMBER = 0x67A40014
    CHILDRENSNAMES = 0x3A58101F
    CLIENTACTIONS = 0x66450102
    CLIENTSUBMITTIME = 0x00390040
    PROVIDERSUBMITTIME = 0x00480040
    CODEPAGEID = 0x66C30003
    COMMENT = 0x3004001F
    COMMENT_STRING8 = 0x3004001E
    COMPANYMAINTELEPHONENUMBER = 0x3A57001F
    COMPANYMAINTELEPHONENUMBER_STRING8 = 0x3A57001E
    COMPANYNAME = 0x3A16001F
    COMPANYNAME_STRING8 = 0x3A16001E
    COMPUTERNETWORKNAME = 0x3A49001F
    COMPUTERNETWORKNAME_STRING8 = 0x3A49001E
    CONFLICTENTRYID = 0x3FF00102
    CONTAINERCLASS = 0x3613001F
    CONTAINERCLASS_STRING8 = 0x3613001E
    CONTAINERCONTENTS = 0x360F000D
    CONTAINERFLAGS = 0x36000003
    CONTAINERHIERARCHY = 0x360E000D
    CONTENTCOUNT = 0x36020003
    CONTENTFILTERSPAMCONFIDENCELEVEL = 0x40760003
    CONTENTUNREADCOUNT = 0x36030003
    CONVERSATIONID = 0x30130102
    CONVERSATIONINDEX = 0x00710102
    CONVERSATIONINDEXTRACKING = 0x3016000B
    CONVERSATIONTOPIC = 0x0070001F
    CONVERSATIONTOPIC_STRING8 = 0x0070001E
    COUNTRY = 0x3A26001F
    COUNTRY_STRING8 = 0x3A26001E
    CREATIONTIME = 0x30070040
    CREATORENTRYID = 0x3FF90102
    CREATORNAME = 0x3FF8001F
    CREATORNAME_STRING8 = 0x3FF8001E
    CUSTOMERID = 0x3A4A001F
    CUSTOMERID_STRING8 = 0x3A4A001E
    DAMBACKPATCHED = 0x6647000B
    DAMORIGINALENTRYID = 0x66460102
    DEFAULTPOSTMESSAGECLASS = 0x36E5001F
    DEFAULTPOSTMESSAGECLASS_STRING8 = 0x36E5001E
    DEFERREDACTIONMESSAGEORIGINALENTRYID = 0x674100FB
    RULEFOLDERFID = 0x67420014
    DEFERREDDELIVERYTIME = 0x000F0040
    DEFERREDSENDNUMBER = 0x3FEB0003
    DEFERREDSENDTIME = 0x3FEF0040
    DEFERREDSENDUNITS = 0x3FEC0003
    DELEGATEFLAGS = 0x686B1003
    DELEGATEDBYRULE = 0x3FE3000B
    DELETEAFTERSUBMIT = 0x0E01000B
    SENTMAILENTRYID = 0x0E0A0102
    DEFAULTSTORE = 0x3400000B
    DELETEDCOUNTTOTAL = 0x670B0003
    DELETEDFOLDERTOTAL = 0x66410003
    DELETEDON = 0x668F0040
    DELIVERTIME = 0x00100040
    DEPARTMENTNAME = 0x3A18001F
    DEPARTMENTNAME_STRING8 = 0x3A18001E
    DEPTH = 0x30050003
    DISPLAYBCC = 0x0E02001F
    DISPLAYBCC_STRING8 = 0x0E02001E
    DISPLAYCC = 0x0E03001F
    DISPLAYCC_STRING8 = 0x0E03001E
    DISPLAYNAME = 0x3001001F
    DISPLAYNAMEPREFIX = 0x3A45001F
    DISPLAYNAMEPREFIX_STRING8 = 0x3A45001E
    DISPLAYNAME_STRING8 = 0x3001001E
    DISPLAYTO = 0x0E04001F
    DISPLAYTO_STRING8 = 0x0E04001E
    DISPLAYTYPE = 0x39000003
    DISPLAYTYPEEX = 0x39050003
    EMAILADDRESS = 0x3003001F
    EMAILADDRESS_STRING8 = 0x3003001E
    ENDDATE = 0x00610040
    ENTRYID = 0x0FFF0102
    EXCEPTIONENDTIME = 0x7FFC0040
    EXCEPTIONREPLACETIME = 0x7FF90040
    EXCEPTIONSTARTTIME = 0x7FFB0040
    EXCHANGENTSECURITYDESCRIPTOR = 0x0E840102
    EXPIRYNUMBER = 0x3FED0003
    EXPIRYTIME = 0x00150040
    EXPIRYUNITS = 0x3FEE0003
    EXTENDEDFOLDERFLAGS = 0x36DA0102
    EXTENDEDRULEMESSAGEACTIONS = 0x0E990102
    EXTENDEDRULEMESSAGECONDITION = 0x0E9A0102
    EXTENDEDRULESIZELIMIT = 0x0E9B0003
    FAXNUMBEROFPAGES = 0x68040003
    FLAGCOMPLETETIME = 0x10910040
    FLAGSTATUS = 0x10900003
    FLATURLNAME = 0x670E001F
    FLATURLNAME_STRING8 = 0x670E001E
    FOLDERASSOCIATEDCONTENTS = 0x3610000D
    FOLDERID = 0x67480014
    FOLDERTYPE = 0x36010003
    FOLLOWUPICON = 0x10950003
    FREEBUSYCOUNTMONTHS = 0x68690003
    FREEBUSYENTRYIDS = 0x36E41102
    FREEBUSYMESSAGEEMAILADDRESS = 0x6849001F
    FREEBUSYMESSAGEEMAILADDRESS_STRING8 = 0x6849001E
    FREEBUSYPUBLISHEND = 0x68480003
    FREEBUSYPUBLISHSTART = 0x68470003
    FREEBUSYRANGETIMESTAMP = 0x68680040
    FTPSITE = 0x3A4C001F
    FTPSITE_STRING8 = 0x3A4C001E
    GATEWAYNEEDSTOREFRESH = 0x6846000B
    GENDER = 0x3A4D0002
    GENERATION = 0x3A05001F
    GENERATION_STRING8 = 0x3A05001E
    GIVENNAME = 0x3A06001F
    GIVENNAME_STRING8 = 0x3A06001E
    GOVERNMENTIDNUMBER = 0x3A07001F
    GOVERNMENTIDNUMBER_STRING8 = 0x3A07001E
    HASATTACHMENTS = 0x0E1B000B
    HASDEFERREDACTIONMESSAGES = 0x3FEA000B
    HASNAMEDPROPERTIES = 0x664A000B
    HASRULES = 0x663A000B
    HIERARCHYCHANGENUMBER = 0x663E0003
    HOBBIES = 0x3A43001F
    HOBBIES_STRING8 = 0x3A43001E
    HOME2TELEPHONENUMBER = 0x3A2F001F
    HOME2TELEPHONENUMBER_STRING8 = 0x3A2F001E
    HOME2TELEPHONENUMBERS = 0x3A2F101F
    HOMEADDRESSCITY = 0x3A59001F
    HOMEADDRESSCITY_STRING8 = 0x3A59001E
    HOMEADDRESSCOUNTRY = 0x3A5A001F
    HOMEADDRESSCOUNTRY_STRING8 = 0x3A5A001E
    HOMEADDRESSPOSTOFFICEBOX = 0x3A5E001F
    HOMEADDRESSPOSTOFFICEBOX_STRING8 = 0x3A5E001E
    HOMEADDRESSPOSTALCODE = 0x3A5B001F
    HOMEADDRESSPOSTALCODE_STRING8 = 0x3A5B001E
    HOMEADDRESSSTATEORPROVINCE = 0x3A5C001F
    HOMEADDRESSSTATEORPROVINCE_STRING8 = 0x3A5C001E
    HOMEADDRESSSTREET = 0x3A5D001F
    HOMEADDRESSSTREET_STRING8 = 0x3A5D001E
    HOMEFAXNUMBER = 0x3A25001F
    HOMEFAXNUMBER_STRING8 = 0x3A25001E
    HOMETELEPHONENUMBER = 0x3A09001F
    HOMETELEPHONENUMBER_STRING8 = 0x3A09001E
    HTML = 0x10130102
    ICALENDARENDTIME = 0x10C40040
    ICALENDARREMINDERNEXTTIME = 0x10CA0040
    ICALENDARSTARTTIME = 0x10C30040
    ICONINDEX = 0x10800003
    IMPORTANCE = 0x00170003
    INCONFLICT = 0x666C000B
    INREPLYTOID = 0x1042001F
    INREPLYTOID_STRING8 = 0x1042001E
    INITIALDETAILSPANE = 0x3F080003
    INITIALS = 0x3A0A001F
    INITIALS_STRING8 = 0x3A0A001E
    INSTID = 0x674D0014
    INSTANCEKEY = 0x0FF60102
    INSTANCESVREID = 0x0FF600FB
    INSTANCENUM = 0x674E0003
    INTERNETCODEPAGE = 0x3FDE0003
    INTERNETMAILOVERRIDEFORMAT = 0x59020003
    INTERNETMESSAGEID = 0x1035001F
    INTERNETMESSAGEID_STRING8 = 0x1035001E
    INTERNETREFERENCES = 0x1039001F
    INTERNETREFERENCES_STRING8 = 0x1039001E
    IPMAPPOINTMENTENTRYID = 0x36D00102
    IPMCONTACTENTRYID = 0x36D10102
    IPMDRAFTSENTRYID = 0x36D70102
    IPMJOURNALENTRYID = 0x36D20102
    IPMNOTEENTRYID = 0x36D30102
    IPMTASKENTRYID = 0x36D40102
    ISDNNUMBER = 0x3A2D001F
    ISDNNUMBER_STRING8 = 0x3A2D001E
    JUNKADDRECIPIENTSTOSAFESENDERSLIST = 0x61030003
    JUNKINCLUDECONTACTS = 0x61000003
    JUNKPERMANENTLYDELETE = 0x61020003
    JUNKPHISHINGENABLELINKS = 0x6107000B
    JUNKTHRESHOLD = 0x61010003
    KEYWORD = 0x3A0B001F
    KEYWORD_STRING8 = 0x3A0B001E
    LANGUAGE = 0x3A0C001F
    LANGUAGE_STRING8 = 0x3A0C001E
    LASTMODIFICATIONTIME = 0x30080040
    LASTMODIFIERENTRYID = 0x3FFB0102
    LASTMODIFIERNAME = 0x3FFA001F
    LASTMODIFIERNAME_STRING8 = 0x3FFA001E
    LASTVERBEXECUTED = 0x10810003
    LASTVERBEXECUTIONTIME = 0x10820040
    LISTHELP = 0x1043001F
    LISTHELP_STRING8 = 0x1043001E
    LISTSUBSCRIBE = 0x1044001F
    LISTSUBSCRIBE_STRING8 = 0x1044001E
    LISTUNSUBSCRIBE = 0x1045001F
    LISTUNSUBSCRIBE_STRING8 = 0x1045001E
    LOCALCOMMITTIME = 0x67090040
    LOCALCOMMITTIMEMAX = 0x670A0040
    LOCALEID = 0x66A10003
    LOCALITY = 0x3A27001F
    LOCALITY_STRING8 = 0x3A27001E
    LOCATION = 0x3A0D001F
    LOCATION_STRING8 = 0x3A0D001E
    MAILBOXOWNERENTRYID = 0x661B0102
    MAILBOXOWNERNAME = 0x661C001F
    MAILBOXOWNERNAME_STRING8 = 0x661C001E
    MANAGERNAME = 0x3A4E001F
    MANAGERNAME_STRING8 = 0x3A4E001E
    MAPPINGSIGNATURE = 0x0FF80102
    MAXIMUMSUBMITMESSAGESIZE = 0x666D0003
    MEMBERID = 0x66710014
    MEMBERNAME = 0x6672001F
    MEMBERNAME_STRING8 = 0x6672001E
    MEMBERRIGHTS = 0x66730003
    MESSAGEATTACHMENTS = 0x0E13000D
    MESSAGECCME = 0x0058000B
    MESSAGECLASS = 0x001A001F
    MESSAGECLASS_STRING8 = 0x001A001E
    MESSAGECODEPAGE = 0x3FFD0003
    MESSAGEDELIVERYTIME = 0x0E060040
    MESSAGEEDITORFORMAT = 0x59090003
    MESSAGEFLAGS = 0x0E070003
    ITEMTEMPORARYFLAGS = 0x10970003
    MESSAGEHANDLINGSYSTEMCOMMONNAME = 0x3A0F001F
    MESSAGEHANDLINGSYSTEMCOMMONNAME_STRING8 = 0x3A0F001E
    MESSAGELOCALEID = 0x3FF10003
    MESSAGERECIPIENTME = 0x0059000B
    MESSAGERECIPIENTS = 0x0E12000D
    MESSAGESIZE = 0x0E080003
    NORMALMESSAGESIZE = 0x66B30003
    NORMALMESSAGESIZEEXTENDED = 0x66B30014
    ASSOCMESSAGESIZE = 0x66B40003
    ASSOCMESSAGESIZEEXTENDED = 0x66B40014
    MESSAGESIZEEXTENDED = 0x0E080014
    MESSAGESTATUS = 0x0E170003
    MESSAGESUBMISSIONID = 0x00470102
    MESSAGETOME = 0x0057000B
    MID = 0x674A0014
    MIDDLENAME = 0x3A44001F
    MIDDLENAME_STRING8 = 0x3A44001E
    MIMESKELETON = 0x64F00102
    MOBILETELEPHONENUMBER = 0x3A1C001F
    MOBILETELEPHONENUMBER_STRING8 = 0x3A1C001E
    NATIVEBODY = 0x10160003
    NEXTSENDACCT = 0x0E29001F
    NEXTSENDACCT_STRING8 = 0x0E29001E
    NICKNAME = 0x3A4F001F
    NICKNAME_STRING8 = 0x3A4F001E
    NONDELIVERYREPORTDIAGCODE = 0x0C050003
    NONDELIVERYREPORTREASONCODE = 0x0C040003
    NONDELIVERYREPORTSTATUSCODE = 0x0C200003
    NORMALIZEDSUBJECT = 0x0E1D001F
    NORMALIZEDSUBJECT_STRING8 = 0x0E1D001E
    OBJECTTYPE = 0x0FFE0003
    OFFICELOCATION = 0x3A19001F
    OFFICELOCATION_STRING8 = 0x3A19001E
    OFFLINEADDRESSBOOKCONTAINERGUID = 0x6802001E
    OFFLINEADDRESSBOOKDISTINGUISHEDNAME = 0x6804001E
    OFFLINEADDRESSBOOKMESSAGECLASS = 0x68030003
    OFFLINEADDRESSBOOKNAME = 0x6800001F
    OFFLINEADDRESSBOOKNAME_STRING8 = 0x6800001E
    OFFLINEADDRESSBOOKSEQUENCE = 0x68010003
    OFFLINEADDRESSBOOKTRUNCATEDPROPERTIES = 0x68051003
    ORDINALMOST = 0x36E20003
    ORGANIZATIONALIDNUMBER = 0x3A10001F
    ORGANIZATIONALIDNUMBER_STRING8 = 0x3A10001E
    ORIGINALAUTHORENTRYID = 0x004C0102
    ORIGINALAUTHORNAME = 0x004D001F
    ORIGINALAUTHORNAME_STRING8 = 0x004D001E
    ORIGINALDELIVERYTIME = 0x00550040
    ORIGINALDISPLAYBCC = 0x0072001F
    ORIGINALDISPLAYBCC_STRING8 = 0x0072001E
    ORIGINALDISPLAYCC = 0x0073001F
    ORIGINALDISPLAYCC_STRING8 = 0x0073001E
    ORIGINALDISPLAYTO = 0x0074001F
    ORIGINALDISPLAYTO_STRING8 = 0x0074001E
    ORIGINALENTRYID = 0x3A120102
    ORIGINALMESSAGECLASS = 0x004B001F
    ORIGINALMESSAGECLASS_STRING8 = 0x004B001E
    ORIGINALMESSAGEID = 0x1046001F
    ORIGINALMESSAGEID_STRING8 = 0x1046001E
    ORIGINALSENDERADDRESSTYPE = 0x0066001F
    ORIGINALSENDERADDRESSTYPE_STRING8 = 0x0066001E
    ORIGINALSENDEREMAILADDRESS = 0x0067001F
    ORIGINALSENDEREMAILADDRESS_STRING8 = 0x0067001E
    ORIGINALSENDERENTRYID = 0x005B0102
    ORIGINALSENDERNAME = 0x005A001F
    ORIGINALSENDERNAME_STRING8 = 0x005A001E
    ORIGINALSENDERSEARCHKEY = 0x005C0102
    ORIGINALSENSITIVITY = 0x002E0003
    ORIGINALSENTREPRESENTINGADDRESSTYPE = 0x0068001F
    ORIGINALSENTREPRESENTINGADDRESSTYPE_STRING8 = 0x0068001E
    ORIGINALSENTREPRESENTINGEMAILADDRESS = 0x0069001F
    ORIGINALSENTREPRESENTINGEMAILADDRESS_STRING8 = 0x0069001E
    ORIGINALSENTREPRESENTINGENTRYID = 0x005E0102
    ORIGINALSENTREPRESENTINGNAME = 0x005D001F
    ORIGINALSENTREPRESENTINGNAME_STRING8 = 0x005D001E
    ORIGINALSENTREPRESENTINGSEARCHKEY = 0x005F0102
    ORIGINALSUBJECT = 0x0049001F
    ORIGINALSUBJECT_STRING8 = 0x0049001E
    ORIGINALSUBMITTIME = 0x004E0040
    ORIGINATORDELIVERYREPORTREQUESTED = 0x0023000B
    ORIGINATORNONDELIVERYREPORTREQUESTED = 0x0C08000B
    OSCSYNCENABLED = 0x7C24000B
    OTHERADDRESSCITY = 0x3A5F001F
    OTHERADDRESSCITY_STRING8 = 0x3A5F001E
    OTHERADDRESSCOUNTRY = 0x3A60001F
    OTHERADDRESSCOUNTRY_STRING8 = 0x3A60001E
    OTHERADDRESSPOSTOFFICEBOX = 0x3A64001F
    OTHERADDRESSPOSTOFFICEBOX_STRING8 = 0x3A64001E
    OTHERADDRESSPOSTALCODE = 0x3A61001F
    OTHERADDRESSPOSTALCODE_STRING8 = 0x3A61001E
    OTHERADDRESSSTATEORPROVINCE = 0x3A62001F
    OTHERADDRESSSTATEORPROVINCE_STRING8 = 0x3A62001E
    OTHERADDRESSSTREET = 0x3A63001F
    OTHERADDRESSSTREET_STRING8 = 0x3A63001E
    OTHERTELEPHONENUMBER = 0x3A1F001F
    OTHERTELEPHONENUMBER_STRING8 = 0x3A1F001E
    OUTOFOFFICESTATE = 0x661D000B
    OWNERAPPOINTMENTID = 0x00620003
    PAGERTELEPHONENUMBER = 0x3A21001F
    PAGERTELEPHONENUMBER_STRING8 = 0x3A21001E
    PARENTENTRYID = 0x0E090102
    PARENTSVREID = 0x0E0900FB
    PARENTFOLDERID = 0x67490014
    PARENTKEY = 0x00250102
    PARENTSOURCEKEY = 0x65E10102
    PERSONALHOMEPAGE = 0x3A50001F
    PERSONALHOMEPAGE_STRING8 = 0x3A50001E
    POLICYTAG = 0x30190102
    POSTOFFICEBOX = 0x3A2B001F
    POSTOFFICEBOX_STRING8 = 0x3A2B001E
    POSTALADDRESS = 0x3A15001F
    POSTALADDRESS_STRING8 = 0x3A15001E
    POSTALCODE = 0x3A2A001F
    POSTALCODE_STRING8 = 0x3A2A001E
    PREDECESSORCHANGELIST = 0x65E30102
    PRIMARYFAXNUMBER = 0x3A23001F
    PRIMARYFAXNUMBER_STRING8 = 0x3A23001E
    PRIMARYSENDACCOUNT = 0x0E28001F
    PRIMARYSENDACCOUNT_STRING8 = 0x0E28001E
    PRIMARYTELEPHONENUMBER = 0x3A1A001F
    PRIMARYTELEPHONENUMBER_STRING8 = 0x3A1A001E
    PRIORITY = 0x00260003
    PROCESSED = 0x7D01000B
    PROFESSION = 0x3A46001F
    PROFESSION_STRING8 = 0x3A46001E
    PROHIBITRECEIVEQUOTA = 0x666A0003
    PROHIBITSENDQUOTA = 0x666E0003
    PROVIDERDISPLAY = 0x3006001F
    PROVIDERDISPLAY_STRING8 = 0x3006001E
    PROVIDERUID = 0x300C0102
    PURPORTEDSENDERDOMAIN = 0x4083001F
    PURPORTEDSENDERDOMAIN_STRING8 = 0x4083001E
    RADIOTELEPHONENUMBER = 0x3A1D001F
    RADIOTELEPHONENUMBER_STRING8 = 0x3A1D001E
    READ = 0x0E69000B
    READRECEIPTADDRESSTYPE = 0x4029001F
    READRECEIPTADDRESSTYPE_STRING8 = 0x4029001E
    READRECEIPTEMAILADDRESS = 0x402A001F
    READRECEIPTEMAILADDRESS_STRING8 = 0x402A001E
    READRECEIPTENTRYID = 0x00460102
    READRECEIPTNAME = 0x402B001F
    READRECEIPTNAME_STRING8 = 0x402B001E
    READRECEIPTREQUESTED = 0x0029000B
    READRECEIPTSEARCHKEY = 0x00530102
    READRECEIPTSMTPADDRESS = 0x5D05001F
    READRECEIPTSMTPADDRESS_STRING8 = 0x5D05001E
    RECEIPTTIME = 0x002A0040
    RECEIVEDBYADDRESSTYPE = 0x0075001F
    RECEIVEDBYADDRESSTYPE_STRING8 = 0x0075001E
    RECEIVEDBYEMAILADDRESS = 0x0076001F
    RECEIVEDBYEMAILADDRESS_STRING8 = 0x0076001E
    RECEIVEDBYENTRYID = 0x003F0102
    RECEIVEDBYNAME = 0x0040001F
    RECEIVEDBYNAME_STRING8 = 0x0040001E
    RECEIVEDBYSEARCHKEY = 0x00510102
    RECEIVEDBYSMTPADDRESS = 0x5D07001F
    RECEIVEDBYSMTPADDRESS_STRING8 = 0x5D07001E
    RECEIVEDREPRESENTINGADDRESSTYPE = 0x0077001F
    RECEIVEDREPRESENTINGADDRESSTYPE_STRING8 = 0x0077001E
    RECEIVEDREPRESENTINGEMAILADDRESS = 0x0078001F
    RECEIVEDREPRESENTINGEMAILADDRESS_STRING8 = 0x0078001E
    RECEIVEDREPRESENTINGENTRYID = 0x00430102
    RECEIVEDREPRESENTINGNAME = 0x0044001F
    RECEIVEDREPRESENTINGNAME_STRING8 = 0x0044001E
    RECEIVEDREPRESENTINGSEARCHKEY = 0x00520102
    RECEIVEDREPRESENTINGSMTPADDRESS = 0x5D08001F
    RECEIVEDREPRESENTINGSMTPADDRESS_STRING8 = 0x5D08001E
    RECIPIENTDISPLAYNAME = 0x5FF6001F
    RECIPIENTDISPLAYNAME_STRING8 = 0x5FF6001E
    RECIPIENTENTRYID = 0x5FF70102
    RECIPIENTFLAGS = 0x5FFD0003
    RECIPIENTORDER = 0x5FDF0003
    RECIPIENTPROPOSED = 0x5FE1000B
    RECIPIENTPROPOSEDENDTIME = 0x5FE40040
    RECIPIENTPROPOSEDSTARTTIME = 0x5FE30040
    RECIPIENTREASSIGNMENTPROHIBITED = 0x002B000B
    RECIPIENTTRACKSTATUS = 0x5FFF0003
    RECIPIENTTRACKSTATUSTIME = 0x5FFB0040
    RECIPIENTTYPE = 0x0C150003
    RECORDKEY = 0x0FF90102
    REFERREDBYNAME = 0x3A47001F
    REFERREDBYNAME_STRING8 = 0x3A47001E
    REMINDERSONLINEENTRYID = 0x36D50102
    REMOTEMESSAGETRANSFERAGENT = 0x0C21001F
    REMOTEMESSAGETRANSFERAGENT_STRING8 = 0x0C21001E
    RENDERINGPOSITION = 0x370B0003
    REPLYRECIPIENTENTRIES = 0x004F0102
    REPLYRECIPIENTNAMES = 0x0050001F
    REPLYRECIPIENTNAMES_STRING8 = 0x0050001E
    REPLYREQUESTED = 0x0C17000B
    REPLYTEMPLATEID = 0x65C20102
    REPLYTIME = 0x00300040
    REPORTDISPOSITION = 0x0080001F
    REPORTDISPOSITIONMODE = 0x0081001F
    REPORTDISPOSITIONMODE_STRING8 = 0x0081001E
    REPORTDISPOSITION_STRING8 = 0x0080001E
    REPORTENTRYID = 0x00450102
    REPORTNAME = 0x003A001F
    REPORTNAME_STRING8 = 0x003A001E
    REPORTSEARCHKEY = 0x00540102
    REPORTTAG = 0x00310102
    REPORTTEXT = 0x1001001F
    REPORTTEXT_STRING8 = 0x1001001E
    REPORTTIME = 0x00320040
    REPORTINGMESSAGETRANSFERAGENT = 0x6820001F
    REPORTINGMESSAGETRANSFERAGENT_STRING8 = 0x6820001E
    RESOLVEMETHOD = 0x3FE70003
    RESOURCEFLAGS = 0x30090003
    RESOURCETYPE = 0x3E030003
    RESPONSEREQUESTED = 0x0063000B
    RESPONSIBILITY = 0x0E0F000B
    RETENTIONDATE = 0x301C0040
    RETENTIONFLAGS = 0x301D0003
    RETENTIONPERIOD = 0x301A0003
    RIGHTS = 0x66390003
    ROAMINGDATATYPES = 0x7C060003
    ROAMINGDICTIONARY = 0x7C070102
    ROAMINGXMLSTREAM = 0x7C080102
    ROWTYPE = 0x0FF50003
    ROWID = 0x30000003
    RTFCOMPRESSED = 0x10090102
    RTFINSYNC = 0x0E1F000B
    RULEACTIONNUMBER = 0x66500003
    RULEACTIONTYPE = 0x66490003
    RULEACTIONS = 0x668000FE
    RULECONDITION = 0x667900FD
    RULEERROR = 0x66480003
    RULEFOLDERENTRYID = 0x66510102
    RULEID = 0x66740014
    RULEIDS = 0x66750102
    RULELEVEL = 0x66830003
    RULEMESSAGELEVEL = 0x65ED0003
    RULEMESSAGENAME = 0x65EC001F
    RULEMESSAGENAME_STRING8 = 0x65EC001E
    RULEMESSAGEPROVIDER = 0x65EB001F
    RULEMESSAGEPROVIDERDATA = 0x65EE0102
    RULEMESSAGEPROVIDER_STRING8 = 0x65EB001E
    RULEMESSAGESEQUENCE = 0x65F30003
    RULEMESSAGESTATE = 0x65E90003
    RULEMESSAGEUSERFLAGS = 0x65EA0003
    RULENAME = 0x6682001F
    RULENAME_STRING8 = 0x6682001E
    RULEPROVIDER = 0x6681001F
    RULEPROVIDERDATA = 0x66840102
    RULEPROVIDER_STRING8 = 0x6681001E
    RULESEQUENCE = 0x66760003
    RULESTATE = 0x66770003
    RULEUSERFLAGS = 0x66780003
    RWRULESSTREAM = 0x68020102
    SCHEDULEINFOAPPOINTMENTTOMBSTONE = 0x686A0102
    SCHEDULEINFOAUTOACCEPTAPPOINTMENTS = 0x686D000B
    SCHEDULEINFODELEGATEENTRYIDS = 0x68451102
    SCHEDULEINFODELEGATENAMES = 0x6844101F
    SCHEDULEINFODELEGATENAMESW = 0x684A101F
    SCHEDULEINFODELEGATORWANTSCOPY = 0x6842000B
    SCHEDULEINFODELEGATORWANTSINFO = 0x684B000B
    SCHEDULEINFODISALLOWOVERLAPPINGAPPTS = 0x686F000B
    SCHEDULEINFODISALLOWRECURRINGAPPTS = 0x686E000B
    SCHEDULEINFODONTMAILDELEGATES = 0x6843000B
    SCHEDULEINFOFREEBUSY = 0x686C0102
    SCHEDULEINFOFREEBUSYAWAY = 0x68561102
    SCHEDULEINFOFREEBUSYBUSY = 0x68541102
    SCHEDULEINFOFREEBUSYMERGED = 0x68501102
    SCHEDULEINFOFREEBUSYTENTATIVE = 0x68521102
    SCHEDULEINFOMONTHSAWAY = 0x68551003
    SCHEDULEINFOMONTHSBUSY = 0x68531003
    SCHEDULEINFOMONTHSMERGED = 0x684F1003
    SCHEDULEINFOMONTHSTENTATIVE = 0x68511003
    SCHEDULEINFORESOURCETYPE = 0x68410003
    SCHEDULEPLUSFREEBUSYENTRYID = 0x66220102
    SCRIPTDATA = 0x00040102
    SEARCHFOLDERDEFINITION = 0x68450102
    SEARCHFOLDEREXPIRATION = 0x683A0003
    SEARCHFOLDERID = 0x68420102
    SEARCHFOLDERLASTUSED = 0x68340003
    SEARCHFOLDERRECREATEINFO = 0x68440102
    SEARCHFOLDERSTORAGETYPE = 0x68460003
    SEARCHKEY = 0x300B0102
    SECURITYDESCRIPTORASXML = 0x0E6A001F
    SECURITYDESCRIPTORASXML_STRING8 = 0x0E6A001E
    SELECTABLE = 0x3609000B
    SENDINTERNETENCODING = 0x3A710003
    SENDRICHINFO = 0x3A40000B
    SENDERADDRESSTYPE = 0x0C1E001F
    SENDERADDRESSTYPE_STRING8 = 0x0C1E001E
    SENDEREMAILADDRESS = 0x0C1F001F
    SENDEREMAILADDRESS_STRING8 = 0x0C1F001E
    SENDERENTRYID = 0x0C190102
    SENDERIDSTATUS = 0x40790003
    SENDERNAME = 0x0C1A001F
    SENDERNAME_STRING8 = 0x0C1A001E
    SENDERSEARCHKEY = 0x0C1D0102
    SENDERSMTPADDRESS = 0x5D01001F
    SENDERSMTPADDRESS_STRING8 = 0x5D01001E
    SENDERTELEPHONENUMBER = 0x6802001F
    SENDERTELEPHONENUMBER_STRING8 = 0x6802001E
    SENSITIVITY = 0x00360003
    SENTMAILSVREID = 0x674000FB
    SENTREPRESENTINGADDRESSTYPE = 0x0064001F
    SENTREPRESENTINGADDRESSTYPE_STRING8 = 0x0064001E
    SENTREPRESENTINGEMAILADDRESS = 0x0065001F
    SENTREPRESENTINGEMAILADDRESS_STRING8 = 0x0065001E
    SENTREPRESENTINGENTRYID = 0x00410102
    SENTREPRESENTINGFLAGS = 0x401A0003
    SENTREPRESENTINGNAME = 0x0042001F
    SENTREPRESENTINGNAME_STRING8 = 0x0042001E
    SENTREPRESENTINGSEARCHKEY = 0x003B0102
    SENTREPRESENTINGSMTPADDRESS = 0x5D02001F
    SENTREPRESENTINGSMTPADDRESS_STRING8 = 0x5D02001E
    SMTPADDRESS = 0x39FE001F
    SMTPADDRESS_STRING8 = 0x39FE001E
    SORTLOCALEID = 0x67050003
    SOURCEKEY = 0x65E00102
    SPOKENNAME = 0x8CC20102
    SPOUSENAME = 0x3A48001F
    SPOUSENAME_STRING8 = 0x3A48001E
    STARTDATE = 0x00600040
    STARTDATEETC = 0x301B0102
    STATEORPROVINCE = 0x3A28001F
    STATEORPROVINCE_STRING8 = 0x3A28001E
    STOREENTRYID = 0x0FFB0102
    STOREPROVIDER = 0x34140102
    STORERECORDKEY = 0x0FFA0102
    STORESTATE = 0x340E0003
    STORESUPPORTMASK = 0x340D0003
    STREETADDRESS = 0x3A29001F
    STREETADDRESS_STRING8 = 0x3A29001E
    SUBFOLDERS = 0x360A000B
    SUBJECT = 0x0037001F
    SUBJECTPREFIX = 0x003D001F
    SUBJECTPREFIX_STRING8 = 0x003D001E
    SUBJECT_STRING8 = 0x0037001E
    SUPPLEMENTARYINFO = 0x0C1B001F
    SUPPLEMENTARYINFO_STRING8 = 0x0C1B001E
    SURNAME = 0x3A11001F
    SURNAME_STRING8 = 0x3A11001E
    SWAPPEDTODODATA = 0x0E2D0102
    SWAPPEDTODOSTORE = 0x0E2C0102
    TARGETENTRYID = 0x30100102
    TELECOMMUNICATIONSDEVICEFORDEAFTELEPHONENUMBER = 0x3A4B001F
    TELECOMMUNICATIONSDEVICEFORDEAFTELEPHONENUMBER_STRING8 = 0x3A4B001E
    TELEXNUMBER = 0x3A2C001F
    TELEXNUMBER_STRING8 = 0x3A2C001E
    TEMPLATEDATA = 0x00010102
    TEMPLATEID = 0x39020102
    TEXTATTACHMENTCHARSET = 0x371B001F
    TEXTATTACHMENTCHARSET_STRING8 = 0x371B001E
    THUMBNAILPHOTO = 0x8C9E0102
    TITLE = 0x3A17001F
    TITLE_STRING8 = 0x3A17001E
    TNEFCORRELATIONKEY = 0x007F0102
    TODOITEMFLAGS = 0x0E2B0003
    TRANSMITTABLEDISPLAYNAME = 0x3A20001F
    TRANSMITTABLEDISPLAYNAME_STRING8 = 0x3A20001E
    TRANSPORTMESSAGEHEADERS = 0x007D001F
    TRANSPORTMESSAGEHEADERS_STRING8 = 0x007D001E
    TRUSTSENDER = 0x0E790003
    USERCERTIFICATE = 0x3A220102
    USERENTRYID = 0x66190102
    USERX509CERTIFICATE = 0x3A701102
    VIEWDESCRIPTORBINARY = 0x70010102
    VIEWDESCRIPTORNAME = 0x7006001F
    VIEWDESCRIPTORNAME_STRING8 = 0x7006001E
    VIEWDESCRIPTORSTRINGS = 0x7002001F
    VIEWDESCRIPTORSTRINGS_STRING8 = 0x7002001E
    VIEWDESCRIPTORVERSION = 0x70070003
    VOICEMESSAGEATTACHMENTORDER = 0x6805001F
    VOICEMESSAGEATTACHMENTORDER_STRING8 = 0x6805001E
    VOICEMESSAGESENDERNAME = 0x6803001F
    VOICEMESSAGESENDERNAME_STRING8 = 0x6803001E
    VOICEMESSAGEDURATION = 0x68010003
    WEDDINGANNIVERSARY = 0x3A410040
    WLINKADDRESSBOOKEID = 0x68540102
    WLINKADDRESSBOOKSTOREEID = 0x68910102
    WLINKCALENDARCOLOR = 0x68530003
    WLINKCLIENTID = 0x68900102
    WLINKENTRYID = 0x684C0102
    WLINKFLAGS = 0x684A0003
    WLINKFOLDERTYPE = 0x684F0102
    WLINKGROUPCLSID = 0x68500102
    WLINKGROUPHEADERID = 0x68420048
    WLINKGROUPNAME = 0x6851001F
    WLINKORDINAL = 0x684B0102
    WLINKROGROUPTYPE = 0x68920003
    WLINKRECORDKEY = 0x684D0102
    WLINKSECTION = 0x68520003
    WLINKSTOREENTRYID = 0x684E0102
    WLINKTYPE = 0x68490003
    ASSOCIATEDCONTENTCOUNT = 0x36170003
    FOLDERCHILDCOUNT = 0x66380003
    IPMPUBLICFOLDERSENTRYID = 0x66310102
    CONVERSATIONKEY = 0x000B0102
    CONTACTEMAILADDRESSES = 0x3A56101F
    GENERATEEXCHANGEVIEWS = 0x36E9000B
    LATESTDELIVERYTIME = 0x00190040
    MAILPERMISSION = 0x3A0E000B
    FOLDERFLAGS = 0x66A80003
    HIERREV = 0x40820040
    SETGIVEN = 0x40170003
    CNSETSEEN = 0x67960102
    CNSETSEENFAI = 0x67da0102
    CNSETREAD = 0x67d20102
    SETDELETED = 0x67e50102
    SETNOLONGERINSCOPE = 0x40210102
    SETEXPIRED = 0x67930102
    SETREAD = 0x402d0102
    SETUNREAD = 0x402e0102
    NONRECEIPTNOTIFICATIONREQUESTED = 0x0C06000B
    RESERVED = 0xFFFFFFFF
    COMMONVIEWSENTRYID = 0x35E60102
    DELETEDASSOCMESSAGESIZE = 0x669D0003
    DELETEDASSOCMESSAGESIZEEXTENDED = 0x669D0014
    DELETEDASSOCMSGCOUNT = 0x66430003
    DELETEDMESSAGESIZE = 0x669B0003
    DELETEDMESSAGESIZEEXTENDED = 0x669B0014
    DELETEDMSGCOUNT = 0x66400003
    DELETEDNORMALMESSAGESIZE = 0x669C0003
    DELETEDNORMALMESSAGESIZEEXTENDED = 0x669C0014
    FINDERENTRYID = 0x35E70102
    INTERNETARTICLENUMBER = 0x0E230003
    MIDSTRING = 0x0E23001F
    ARTICLENUMBERNEXT = 0x67510003
    IPMDAFENTRYID = 0x661F0102
    IPMFAVORITESENTRYID = 0x66300102
    IPMOUTBOXENTRYID = 0x35E20102
    IPMSENTMAILENTRYID = 0x35E40102
    IPMSUBTREEENTRYID = 0x35E00102
    IPMWASTEBASKETENTRYID = 0x35E30102
    SCHEDULEFOLDERENTRYID = 0x661E0102
    NONIPMSUBTREEENTRYID = 0x66200102
    STORAGEQUOTALIMIT = 0x3FF50003
    STOREOFFLINE = 0x6632000B
    VALIDFOLDERMASK = 0x35DF0003
    VIEWSENTRYID = 0x35E50102
    EFORMSREGISTRYENTRYID = 0x66210102
    OFFLINEADDRBOOKENTRYID = 0x66230102
    HIERARCHYSERVER = 0x6633001E
    REPLICASERVER = 0x6644001E
    FOLDERPATHNAME = 0x66B5001F
    ASSOCMSGATTACHCOUNT = 0x66AE0003
    NORMALMSGATTACHCOUNT = 0x66AD0003
    TESTLINESPEED = 0x662B0102
    PARENTDISPLAY = 0x0E05001F
    PARENTDISPLAY_STRING8 = 0x0E05001E
    CREATORSID = 0x0E580102

    sizeTags = {ASSOCMESSAGESIZE, ASSOCMESSAGESIZEEXTENDED, ATTACHSIZE, DELETEDASSOCMESSAGESIZE,
                DELETEDASSOCMESSAGESIZEEXTENDED, DELETEDMESSAGESIZE, DELETEDMESSAGESIZEEXTENDED, DELETEDNORMALMESSAGESIZE,
                DELETEDNORMALMESSAGESIZEEXTENDED, EXTENDEDRULESIZELIMIT, MAXIMUMSUBMITMESSAGESIZE, MESSAGESIZE,
                MESSAGESIZEEXTENDED, NORMALMESSAGESIZE, NORMALMESSAGESIZEEXTENDED, STORAGEQUOTALIMIT, PROHIBITRECEIVEQUOTA,
                PROHIBITSENDQUOTA}

    sizeFactor = {PROHIBITRECEIVEQUOTA: 1024, PROHIBITSENDQUOTA: 1024, STORAGEQUOTALIMIT: 1024}

    @classmethod
    def deriveTag(cls, tag):
        """Derive numeric tag value from integer or string.

        Parameters
        ----------
        tag : int or str
            - Integer ID of the tag
            - Name of the tag (case insensitive)
            - str that can be converted by int(tag, 0)

        Raises
        ------
        ValueError
            Tag derivation failed

        Returns
        -------
        int
            Numeric tag value
        """
        try:
            return tag if isinstance(tag, int) else getattr(cls, tag.upper(), None) or int(tag, 0)
        except Exception:
            pass
        raise ValueError("Failed to derive proptag from {}".format(repr(tag)))

    @classmethod
    def convertValue(cls, tag, value):
        """Convert value to the appropriate tag type.

        Parameters
        ----------
        tag : int
            Tag ID
        value : any
            Value to convert

        Raises
        ------
        ValueError
            Value cannot be converted to correct type

        Returns
        -------
        value : any
            Converted value
        """
        tagtype = tag & 0xFFFF
        baseType = tag & 0x0FFF
        if tagtype == PropTypes.FILETIME:
            from datetime import datetime
            from .rop import ntTime
            import time
            if not isinstance(value, int):
                if not isinstance(value, datetime):
                    try:
                        value = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
                    except TypeError:
                        raise ValueError("Invalid date '{}'".format(value))
                value = ntTime(time.mktime(value.timetuple()))
        if type(value) != PropTypes.pyType(baseType):
            try:
                value = PropTypes.pyType(baseType)(value)
            except Exception:
                raise ValueError("Type of value {} does not match type of tag {} ({})"
                                 .format(value, cls.lookup(tag), PropTypes.lookup(tag)))
        if tag == cls.DISPLAYTYPEEX:
            value = value & ~0x40000000  # Remove DTE_FLAG_ACL_CAPABLE flag because it is not recognized properly by gromox
        return value

    @classmethod
    def normalize(cls, tag, value):
        """Normalize tag ID and value.

        Calls deriveTag and convertValue.

        Parameters
        ----------
        tag : int or str
            See deriveTag.
        value : any
            Tag value

        Returns
        -------
        value : tuple
            tag ID and value
        """
        tag = cls.deriveTag(tag)
        value = cls.convertValue(tag, value)
        return tag, value


class ConfigIDs(_ReverseLookup):
    MAILBOX_GUID = 1
    CURRENT_EID = 2
    MAXIMUM_EID = 3
    LAST_CHANGE_NUMBER = 4
    LAST_ARTICLE_NUMBER = 5
    LAST_CID = 6
    SEARCH_STATE = 7
    DEFAULT_PERMISSION = 8
    ANONYMOUS_PERMISSION = 9


class PublicFIDs(_ReverseLookup):
    ROOT = 0x01
    IPMSUBTREE = 0x02
    NONIPMSUBTREE = 0x03
    EFORMSREGISTRY = 0x04
    CUSTOM = 0x05


class PrivateFIDs(_ReverseLookup):
    ROOT = 0x01
    DEFERRED_ACTION = 0x02
    SPOOLER_QUEUE = 0x03
    SHORTCUTS = 0x04
    FINDER = 0x05
    VIEWS = 0x06
    COMMON_VIEWS = 0x07
    SCHEDULE = 0x08
    IPMSUBTREE = 0x09
    SENT_ITEMS = 0x0a
    DELETED_ITEMS = 0x0b
    OUTBOX = 0x0c
    INBOX = 0x0d
    DRAFT = 0x0e
    CALENDAR = 0x0f
    JOURNAL = 0x10
    NOTES = 0x11
    TASKS = 0x12
    CONTACTS = 0x13
    QUICKCONTACTS = 0x14
    IMCONTACTLIST = 0x15
    GALCONTACTS = 0x16
    JUNK = 0x17
    LOCAL_FREEBUSY = 0x18
    SYNC_ISSUES = 0x19
    CONFLICTS = 0x1a
    LOCAL_FAILURES = 0x1b
    SERVER_FAILURES = 0x1c
    CONVERSATION_ACTION_SETTINGS = 0x1d
    CUSTOM = 0x1e


class Permissions:
    NONE = 0x00000000
    READANY = 0x00000001
    CREATE = 0x00000002
    SENDAS = 0x00000004  # self defined
    EDITOWNED = 0x00000008
    DELETEOWNED = 0x00000010
    EDITANY = 0x00000020
    DELETEANY = 0x00000040
    CREATESUBFOLDER = 0x00000080
    FOLDEROWNER = 0x00000100
    FOLDERCONTACT = 0x00000200
    FOLDERVISIBLE = 0x00000400
    FREEBUSYSIMPLE = 0x00000800
    FREEBUSYDETAILED = 0x00001000
    STOREOWNER = 0x00002000

    @classmethod
    def domainDefault(cls):
        return cls.READANY | cls.CREATE | cls.FOLDERVISIBLE | cls.EDITOWNED | cls.DELETEOWNED


class TableFlags(_ReverseLookup):
    ASSOCIATED = 0x2
    DEPTH = 0x4
    DEFERREDERRORS = 0x8
    NONOTIFICATIONS = 0x10
    SOFTDELETES = 0x20
    USEUNICODE = 0x40
    SUPPRESSNOTIFICATIONS = 0x80
    CONVERSATIONMEMBERS = 0x80


class Misc:
    ALLOCATED_EID_RANGE = 0x10000
    CHANGE_NUMBER_BEGIN = 0x800000000000


class FolderNames():
    @classmethod
    def _load(cls):
        if cls._lang is not None:
            return
        try:
            import yaml
            with open("res/foldernames.yaml", encoding="utf-8") as file:
                cls._lang = yaml.load(file, Loader=yaml.SafeLoader)
        except Exception:
            pass

    @classmethod
    def get(cls, name, lang):
        cls._load()
        return cls._lang.get(lang, {}).get(name, getattr(cls, "_"+name)) if cls._lang is not None else getattr(cls, "_"+name)

    _lang = None

    _IPM = "Top of Information Store"
    _INBOX = "Inbox"
    _DRAFT = "Drafts"
    _OUTBOX = "Outbox"
    _SENT = "Sent Items"
    _DELETED = "Deleted Items"
    _CONTACTS = "Contacts"
    _CALENDAR = "Calendar"
    _JOURNAL = "Journal"
    _NOTES = "Notes"
    _TASKS = "Tasks"
    _JUNK = "Junk E-mail"
    _SYNC = "Sync Issues"
    _CONFLICT = "Conflicts"
    _LOCAL = "Local Failures"
    _SERVER = "Server Failures"



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import ServiceHub


def exmdbHandleException(service, error):
    if isinstance(error, ExmdbService.ConnectionError):
        return ServiceHub.UNAVAILABLE, error.args[0]


@ServiceHub.register("exmdb", exmdbHandleException)
class ExmdbService:
    class _BoundClient:
        def __init__(self, exmdb, host, port, homedir, isPrivate):
            self.__homedir = homedir
            self.__client = exmdb.ExmdbQueries(host, port, homedir, isPrivate)

        def __getattr__(self, attr):
            target = getattr(self.__client, attr)
            if callable(target):
                return lambda *args, **kwargs: target(self.__homedir, *args, **kwargs)
            return target

    __loaded = False
    __symbols = ("ConnectionError", "ExmdbError", "ExmdbProtocolError", "SerializationError", "ExmdbQueries", "Folder",
                 "GUID", "PropertyName", "Restriction")
    __methods = ("TaggedPropval", "FolderList", "FolderMemberList")

    def __init__(self):
        self._loadPyexmdb()
        for method in self.__methods:
            setattr(self, method, getattr(self.pyexmdb, method))

    @classmethod
    def _loadPyexmdb(cls):
        if cls.__loaded:
            return
        try:
            from tools.config import Config
            import pyexmdb
            for symbol in cls.__symbols:
                setattr(cls, symbol, getattr(pyexmdb, symbol))
        except Exception:
            for symbol in cls.__symbols:
                setattr(cls, symbol, type(None))
            raise

        cls.host = Config["options"].get("exmdbHost", "::1")
        cls.port = Config["options"].get("exmdbPort", "5000")
        cls.pyexmdb = pyexmdb
        cls.loaded = True

    def client(self, homedir, isPrivate):
        """Shortcut for creating a client.

        Parameters
        ----------
        homedir : str
            Home directory of the user or domain.
        isPrivate : bool
            Whether it is a user (True) or domain (False) database

        Returns
        -------
        pyexmdb.ExmdbQueries
            Exmdb client
        """
        return self.ExmdbQueries(self.host, self.port, homedir, isPrivate)

    def user(self, user):
        """Create client for user.

        Connect to the users homeserver if specified, otherwise connect to configured exmdbHost.
        Connection is always established using the exmdbPort from the configuration.

        Parameters
        ----------
        user : orm.users.Users
            User object providing homeserver and maildir.

        Returns
        -------
        services.exmdb.ExmdbService._BoundClient
            Exmdb client bound to the specific user
        """
        host = user.homeserver.hostname if user.homeserver is not None else self.host
        return self._BoundClient(self, host, self.port, user.maildir, True)

    def domain(self, domain):
        """Create client for domain.

        Connect to the domains homeserver if specified, otherwise connect to configured exmdbHost.
        Connection is always established using the exmdbPort from the configuration.

        Parameters
        ----------
        domain : orm.domains.Domain
            Domain object providing homeserver and homedir.

        Returns
        -------
        services.exmdb.ExmdbService._BoundClient
            Exmdb client bound to the specific domain
        """
        host = domain.homeserver.hostname if domain.homeserver is not None else self.host
        return self._BoundClient(self, host, self.port, domain.homedir, False)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

__all__ = ["domains", "misc", "users", "ext"]

from sqlalchemy import create_engine, event, select, text
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker, class_mapper, Query, column_property

from tools.config import Config

import logging
logger = logging.getLogger("mysql")


class DBConn:
    def __init__(self, URI):
        import threading
        self.engine = create_engine(URI, pool_recycle=Config["DB"]["sessionTimout"])
        self.session = scoped_session(sessionmaker(self.engine), threading.get_ident)
        self.__version = None
        self.__maxversion = 0
        self.initVersion()

    def __reinit(self):
        outerself = self

        class QueryProperty:
            def __get__(self, obj, type):
                mapper = class_mapper(type)
                return Query(mapper, session=outerself.session()) if mapper is not None else None

        class Model:
            query = QueryProperty()

        self.Base = declarative_base(cls=Model)

    def enableFlask(self, API):
        @API.teardown_appcontext
        def removeSession(*args, **kwargs):
            self.session.remove()

    def testConnection(self, verbose=False):
        try:
            self.session.execute("SELECT 1 FROM DUAL")
        except OperationalError as err:
            self.session.remove()
            return "Database connection failed with error {}: {}".format(err.orig.args[0], err.orig.args[1])
        self.session.remove()

    def _fetchVersion(self, verbose=False):
        """Try to fetch schema version from database.

        Parameters
        ----------
        verbose : bool, optional
            Print status information to log. The default is False.

        Returns
        -------
        version : int
            Version number or None on failure
        """
        try:
            version = int(self.session.execute("SELECT `value` FROM `options` WHERE `key` = 'schemaversion'").fetchone()[0])
            if verbose:
                logger.info("Detected database schema version n"+str(version))
            return version
        except Exception:
            if verbose:
                logger.warning("Failed to detect schema version, assuming up-to-date schema")

    def initVersion(self):
        self.__version = self._fetchVersion(True)
        self.__reinit()

    def requireReload(self):
        """Check if a schema version update is available.

        Only queries the database if current version is undefined or is lower
        than the highest known version (i.e. an update would have an actual effect).

        Returns
        -------
        bool
            Whether an update is available and the schema should be reloaded
        """
        return (self.__version is None or self.__version < self.__maxversion) and self._fetchVersion(False) != self.__version

    @property
    def version(self):
        """Get schema version currently in use.

        Returns
        -------
        int
            Schema version number or None if undefined
        """
        return self.__version

    def minVersion(self, v):
        """Check if schema version is at least `v`.

        If the schema version could not be determined, the most recent version
        is assumed and the check always passes.

        Parameters
        ----------
        v : int
            Required version

        Returns
        -------
        bool
            Whether the required version is satisfied
        """
        def sign(x):
            return -1 if x < 0 else 1 if x > 0 else 0
        self.__maxversion = max(abs(v), self.__maxversion)
        return self.__version is not None and sign(v)*self.__version >= v


def _loadDBConfig():
    """Load database parameters from configuration.

    Outputs error messages to log stream if mandatory parameters are missing.

    Returns
    -------
    str
        Database URI including connection and credentials

    """
    if "DB" not in Config:
        logger.error("No database configuration found")
        return None
    DBconf = Config["DB"]
    if "user" not in DBconf or "pass" not in DBconf:
        logger.error("Database user or password missing")
        return None
    if "database" not in DBconf:
        logger.error("No database specified.")
        return None
    if "host" not in DBconf and "port" not in DBconf:
        logger.info("Database connection not specified. Using default '127.0.0.1:3306'")
    host = DBconf.get("host", "127.0.0.1")
    port = DBconf.get("port", 3306)
    return "mysql+mysqldb://{user}:{password}@{host}:{port}/{db}?charset=utf8mb4".format(user=DBconf["user"],
                                                                                         password=DBconf["pass"],
                                                                                         host=host,
                                                                                         port=port,
                                                                                         db=DBconf["database"])


if Config["options"]["disableDB"]:
    DB = None
    logger.warning("Database disabled in configuration")
else:
    DB_uri = _loadDBConfig()
    if DB_uri is not None:
        DB = DBConn(DB_uri)
        err = DB.testConnection(verbose=True)
        if err is not None:
            logger.warning(err)
    else:
        logger.warning("Database configuration failed. No data will be available")
        DB = None


class Stub:
    def __init__(self, value):
        self.__value__ = value

    def __get__(self, *args):
        return self.__value__

    def __set__(self, *args):
        pass


def OptionalNC(version, default, column):
    """Non-column optional attribute wrapper.

    If the specified version is not reached, a Stub is generated that always
    returns the default value and cannot be modified.

    Parameters
    ----------
    version : int
        Minimum schema version
    default : Any
        Default attribute value to return if version check fails
    column : Any
        Column definition to return if version check passes
    """
    return column if DB.minVersion(version) else Stub(default)


def OptionalC(version, default, column):
    """Column optional attribute wrapper.

    If the specified version is not reached, a column_property is created that
    returns the default value and will not emit any SQL when changed and
    committed.
    In contrast to the non-column version `OptionalNC`, the stub can be used
    like an actual column.

    Parameters
    ----------
    version : int
        Minimum schema version
    default : str
        Default SQL expression to use if version check fails
    column : Any
        Column definition to return if version check passes
    """
    return column if DB.minVersion(version) else column_property(select([text(default)]).as_scalar())


class NotifyTable:
    """Helper class tracking inserts and deletes.

    Automatically calls derived classes `_commit` method to react accordingly"""
    __changed = False
    __active = True

    @classmethod
    def NTtouch(cls, *args, **kwargs):
        """Mark table as changed."""
        cls.__changed = True

    @classmethod
    def NTclear(cls, *args, **kwargs):
        """Mark table as unchanged."""
        cls.__changed = False

    @classmethod
    def NTcommit(cls, *args, **kwargs):
        """Call `_commit` if tbale was changed and tracking is active."""
        if cls.__active and cls.__changed and hasattr(cls, "_commit"):
            cls._commit(*args, **kwargs)
            cls.__changed = False

    @classmethod
    def NTregister(cls):
        """Register SQLAlchemy event handlers."""
        event.listen(cls, "after_delete", cls.NTtouch)
        event.listen(cls, "after_insert", cls.NTtouch)
        event.listen(DB.session, "after_commit", cls.NTcommit)
        event.listen(DB.session, "after_rollback", cls.NTclear)

    @classmethod
    def NTactive(cls, state, clear=False):
        """(De-)activate tracking, optionally clearing state."""
        cls.__active = state
        if clear:
            cls.NTclear()
